-- #import Framework
-- #import ExtAC

-- one hypothetical extensional set
X : Set
(==) : Rel X
Equiv : EquivRel (==)

eqXref = Equiv.1
eqXsym = Equiv.2.1
eqXtra = Equiv.2.2

Pow : Set
Pow = { subset : X -> Bool; 
        (a,b:X) -> a==b -> T (subset a) -> T (subset b)
      }

nonempty : (U:Pow) -> Set
nonempty U = Exists (\a -> T (U.1 a))

Pow' : Set
Pow' = { U : Pow; nonempty U}

el : X -> Pow -> Set
el a U = T (U.1 a)

el' : X -> Pow' -> Set
el' a U = el a U.1

sub : (U,V:Pow) -> Set
sub U V = (x:X) -> el x U -> el x V

equal : (U,V:Pow) -> Set
equal U V = sub U V /\ sub V U

sub' : (U,V:Pow') -> Set
sub' U V = sub U.1 V.1

equal' : (U,V:Pow') -> Set
equal' U V = equal U.1 V.1

forall : Pow -> Set -> Set
forall U P = (a:X) -> el a U -> P

exists : Pow -> Set -> Set
exists U P = Exists (\a -> el a U /\ P)

IntAC (A,B:Set) : (R:A->B->Set) -> 
    ((x:A) -> Exists (\y -> R x y)) -> Exists (\f -> (x:A) -> R x (f x))
IntAC R h = (\a -> (h a).1,  \a -> (h a).2)

ExtPred : Type
ExtPred = { P : Pred X;
            (x,y:X) -> x == y -> P x -> P y }

abstract
  compr : Pred X -> Pow
  compr P = 
      let P' : Pred X
	  P' x = P x /\ ((y:X) -> y == x -> P y)
	  R : X -> Bool -> Set
	  R x b = T b <-> P' x
	  lem1 : (x:X) -> Exists (\b -> R x b)
	  lem1 x = orE (LEM (P' x) ) 
		   (\h -> (true, (\h' -> h, \h' -> tt)))
		   (\h -> (false, (\h' -> whenFalse (P' x) h', \h' -> h h')))
	  choicef : Exists (\f -> (x:X) -> R x (f x))
	  choicef = IntAC R lem1
      in (choicef.1,
	 \a b h h' -> 
	     let lem0 : P' a
		 lem0 = (choicef.2 a).1 h'
		 lem1 : P b
		 lem1 = lem0.2 b (eqXsym a b h)
		 lem2 : R b (choicef.1 b)
		 lem2 = choicef.2 b
		 lem3 : P' b
		 lem3 = (lem1,
			 (\y h0 -> lem0.2 y 
				   (eqXtra 
				     y b a h0 (eqXsym a b h))))
	     in lem2.2 lem3 
	 )

  -- znaci trebav izgleda ova dole da go definiram na ExtPred..

  -- every member of a subset created by comprehension
  -- satisfies the predicate, and vice-versa
  comprLem (P:Pred X) : 
	       (a:X) -> el a (compr P) <-> 
			    (P a /\ ((y:X) -> y==a -> P y))
  comprLem {P} a =
      let P' : Pred X
	  P' x = P x /\ ((y:X) -> y==x -> P y)
	  R : X -> Bool -> Set
	  R x b = T b <-> P' x
	  lem1 : (x:X) -> Exists (\b -> R x b)
	  lem1 x = orE (LEM (P' x) ) 
		   (\h -> (true, (\h' -> h, \h' -> tt)))
		   (\h -> (false, (\h' -> whenFalse (P' x) h', \h' -> h h')))
	  choicef : Exists (\f -> (x:X) -> R x (f x))
	  choicef = IntAC R lem1
      in choicef.2 a

O : Pow
O = compr (\x -> False)

C : Pow -> Pow
C U = compr (\x -> Not (el x U) )

(-) : Pow -> Pow -> Pow
U - V = compr (\x -> (el x U) /\ Not (el x V) )

(*) : Pow -> Pow -> Pow
U * V = compr (\x -> (el x U) /\ (el x V) )

(+) : Pow -> Pow -> Pow
U + V = compr (\x -> (el x U) \/ (el x V) )

Singl : X -> Pow
Singl a = compr (\x -> x==a )

trich : Pow -> Rel X -> Set
trich U (<) = (a,b:X) -> el a U -> el b U -> 
                  ((a<b) <-> (Not (b<a) /\ Not (a==b))) /\
                  (a==b <-> (Not (a<b) /\ Not (b<a)))

trans : Pow -> Rel X -> Set
trans U (<) = (a,b,c:X) -> el a U -> el b U -> el c U ->
                  a<b -> b<c -> a<c

linear : Pow -> Rel X -> Set
linear U (<) = trich U (<) /\ trans U (<)

hasLeast : Pow -> Rel X -> Set
hasLeast U (<)= Exists (\a -> el a U /\ ((b:X) -> el b U -> Not (b<a)))

wellOrdered : Pow -> Rel X -> Set
wellOrdered U (<) = 
    (linear U (<)) /\ ((V:Pow') -> sub V.1 U -> hasLeast V.1 (<))

wellFounded : Pow -> Rel X -> (P:ExtPred) -> Set
wellFounded U (<) P = 
    ((x:X) -> el x U -> ((y:X) -> el y U -> y<x -> P.1 y) -> P.1 x) ->
        ((x:X) -> el x U -> P.1 x)

minimal : (U:Pow)->(R:Rel X)->(V:Pow') -> -- explicit arguments because of al
          wellOrdered U R -> sub V.1 U -> X
minimal U R V hWO hsub = (hWO.2 V hsub).1 

minimalLem : (U:Pow)->(R:Rel X)->(V:Pow')-> -- explicit arguments because of al
             (hWO:wellOrdered U R) -> (hsub:sub V.1 U) -> 
             el (minimal U R V hWO hsub) V.1
minimalLem U R V hWO hsub = (hWO.2 V hsub).2.1

minimalLem2 : (U:Pow)->(R:Rel X)->(V:Pow')-> --explicit arguments because of al
                (hWO:wellOrdered U R) -> (hsub:sub V.1 U) -> 
                (b:X) -> el b V.1 -> Not (R b (minimal U R V hWO hsub))
minimalLem2 U R V hWO hsub = (hWO.2 V hsub).2.2


notnot (A:Set) : Not (Not A) -> A
notnot {A} h = orE (LEM A)
               (\h' -> h')
               (\h' -> whenFalse A (h h'))

notnot' (A:Set) : A -> Not (Not A)
notnot' a = \h -> h a 

notnotimp (A,B:Set) : (Not (Not A) -> Not (Not B)) -> A -> B
notnotimp h a = notnot (h (notnot' a)) 

notImp (P,Q:Set) : Not (P->Q) -> P /\ Not Q
notImp {P} {Q} h = orE (LEM P )
                   (orE (LEM Q )
                    (\q p -> (p, (whenFalse (Not Q) (h (\p' -> q)))) )
                    (\nq p -> (p, nq) ))
                   (orE (LEM Q )
                    (\q np -> whenFalse (P /\ Not Q) (h (\p' -> q)) )
                    (\nq np -> ((whenFalse P
                                 (h (notnotimp (contrap (\nq' -> np))))), nq)))

notImp' (P,Q,S:Set) : Not (P->Q->S) -> P /\ Q /\ Not S
notImp' {P} {Q} {S}  h = 
    let lem1 : P /\ Not (Q->S)
        lem1 = notImp h
        lem2 : P /\ (Q /\ Not S)
        lem2 = (lem1.1, notImp lem1.2)
    in ((lem2.1, lem2.2.1), lem2.2.2)

notex2all (A:Set)(P:Pred A) : Not (Exists (\x -> P x)) -> ((x:A) -> Not (P x))
notex2all {A}{P} h x Px = h (x, Px)

notall2ex (A:Set)(P:Pred A) : Not ((x:A) -> P x) -> Exists (\x -> Not (P x))
notall2ex {A}{P} h =
    let lem1 : Not (Exists (\x -> Not (P x))) -> ((x:A) -> Not (Not (P x)))
        lem1 nex = notex2all nex
        lem2 : ((x:A) -> Not (Not (P x))) -> ((x:A) -> P x)
        lem2 h' x = notnot (h' x)
    in orE (LEM (Exists (\x -> Not (P x))))
           (\l -> l)
           (\r -> whenFalse (Exists (\x -> Not (P x))) (h (lem2 (lem1 r))))

notall2ex' (A:Set)(P,Q:Pred A) : Not ((x:A) -> P x -> Q x) -> 
                                 Exists (\x -> P x /\ Not (Q x))
notall2ex' {A}{P}{Q} h = 
    let lem1 : Exists (\x -> Not (P x -> Q x))
        lem1 = notall2ex h
    in (lem1.1, notImp lem1.2)

notall2ex'' (A:Set)(M,N,P,Q:Pred A) : 
            Not ((x:A) -> M x -> N x -> P x -> Q x) -> 
                Exists (\x -> M x /\ N x /\ P x /\ Not (Q x))
notall2ex'' {A}{M}{N}{P}{Q} h = 
    let lem1 : Exists (\x -> Not (M x -> N x -> P x -> Q x))
        lem1 = notall2ex h
        x : A
        x = lem1.1
        lem2 : Not (M x -> N x -> P x -> Q x) -> M x /\ Not (N x -> P x -> Q x)
        lem2 h' = notImp h'
        lem3 : M x /\ Not (N x -> P x -> Q x) -> M x /\ (N x /\ Not (P x -> Q x))
        lem3 h' = (h'.1, notImp h'.2)
        lem4 : M x /\ (N x /\ Not (P x -> Q x)) -> M x /\ (N x /\ (P x /\ Not (Q x)))
        lem4 h' = (h'.1, (h'.2.1, notImp h'.2.2))
        lem5 : M x /\ (N x /\ (P x /\ Not (Q x))) -> M x /\ N x /\ P x /\ Not (Q x)
        lem5 h' = (((h'.1, h'.2.1), h'.2.2.1), h'.2.2.2)
    in (x, lem5 (lem4 (lem3 (lem2 lem1.2))))

ex2ex (A:Set)(P,Q:Pred A) : Exists P -> ((x:A) -> P x -> Q x) -> Exists Q
ex2ex h1 h2 = (h1.1, h2 h1.1 h1.2)

WO2WF : (U:Pow)->(R:Rel X)->(P:ExtPred) -> wellOrdered U R -> wellFounded U R P
WO2WF {U}{R}{P} hWO 
    = orE (LEM (wellFounded U R P) )
      (\h -> h )
      (\h -> let V : Pow
                 V = compr (\x -> el x U /\ Not (P.1 x))
                 lem1 : sub V U
                 lem1 = \x h -> ((comprLem x).1 h).1.1
                 lem1' : sub U U
                 lem1' = \x h -> h
                 neV : nonempty V
                 neV = let lem1 : Not ((x:X) -> el x U -> P.1 x)
                           lem1 = (notImp h).2
                           lem2 : Exists (\x -> el x U /\ Not (P.1 x))
                           lem2 = notall2ex' lem1
                           v = lem2.1
                           vU = lem2.2.1
                           NotPv = lem2.2.2
                           lem3 : (w:X) -> w==v -> el w U /\ Not (P.1 w)
                           lem3 w eqwv = 
                               (U.2 v w (eqXsym w v eqwv) vU
                               ,(contrap (P.2 w v eqwv)) NotPv)
                           vV : el v V
                           vV = (comprLem v).2 (lem2.2, lem3)
                       in (v, vV)
                 V' : Pow'
                 V' = (V,neV)
                 -- t is the minimal of V
                 t : X 
                 t = minimal U R V' hWO lem1
                 lem2 : el t V
                 lem2 = minimalLem U R V' hWO lem1
                 lem2' : el t U
                 lem2' = lem1 t lem2
                 lem3 : (y:X) -> el y U -> R y t -> P.1 y
                 lem3 = \y hUy hRyt ->
                   orE (LEM (P.1 y))
                     (\h -> h)
                     (\h ->
                      let ext : (z:X) -> z==y -> el z U /\ Not (P.1 z)
                          ext z heq = 
                              (U.2 y z(eqXsym z y heq) hUy
                              ,contrap (P.2 z y heq) h)
                          lem31 : el y V
                          lem31 = (comprLem y).2 ((hUy,h),ext)
                          lem32 : Not (R y t)
                          lem32 = minimalLem2 U R V' hWO lem1 y lem31
                          absurd : False
                          absurd = lem32 hRyt
                      in whenFalse (P.1 y) absurd)
                 lem4 : Not (P.1 t)
                 lem4 = ((comprLem t).1 lem2).1.2
                 lem5 : P.1 t
                 lem5 = (notImp h).1 t lem2' lem3
             in whenFalse (wellFounded U R P) (lem4 lem5))

AC2gamma : Exists (\f -> 
                   ((U:Pow') -> el' (f U) U) /\
                   ((U,V:Pow') -> equal' U V -> f U == f V))
AC2gamma
    = let R : Pow' -> X -> Set
          R = \U x -> el' x U 
          equal'Equiv : EquivRel equal'
          equal'Equiv = 
              let isRefl : Reflexive equal'
                  isRefl  = \x -> (\x h -> h , \x h -> h)
                  isSymm : Symmetrical equal'
                  isSymm  = \x1 x2 h -> (h.2, h.1)
                  isTrans : Transitive equal'
                  isTrans  = \x1 x2 x3 h1 h2 -> 
                             (\x h -> h2.1 x (h1.1 x h),
                              \x h -> h1.2 x (h2.2 x h))
              in (isRefl, isSymm, isTrans)
          lem1 : (U,V:Pow')-> equal' U V-> (z:X)-> el' z U -> el' z V
          lem1 U V heq = heq.1 
          lem2 : (x,y:X) -> x==y -> (W:Pow') -> el' x W -> el' y W
          lem2 x y heq W h = W.1.2 x y heq h 
          lem3 : (U:Pow') -> Exists (\y -> el' y U)
          lem3 U = U.2 
      in ExtAC R 
               equal'
               (==) 
               equal'Equiv 
               Equiv 
               lem1 
               lem2 
               lem3 

gamma : Pow' -> X
gamma = AC2gamma.1


IS : Pow -> X -> Rel X -> Pow
IS U a (<) = compr (\x -> el x U /\ (x<a))

GS : Pow -> Rel X -> Set
GS U R = wellOrdered U R /\ 
             ((a:X) -> el a U -> (ne:nonempty (C (IS U a R))) -> 
                       a == (gamma (C (IS U a R), ne)))

notand (P,Q:Set) : Not (Not P /\ Not Q) <-> P \/ Q
notand {P}{Q} =
    let lem1 : Not (Not P /\ Not Q) -> P \/ Q
        lem1 h = orE (LEM P) 
                 (\p -> orI1 p) 
                 (\np -> orE (LEM Q)
                         (\q -> orI2 q) 
                         (\nq -> whenFalse (P\/Q) (h (np,nq))))
        lem2 : P \/ Q -> Not (Not P /\ Not Q)
        lem2 porq npandnq = orE porq (\p -> npandnq.1 p) (\q -> npandnq.2 q)
    in (lem1, lem2)

notand' (P,Q:Set) : Not (Not P /\ Not Q) -> Q \/ P
notand' h = orE (notand.1 h) (\p -> orI2 p) (\q -> orI1 q)

notand'' (P,Q:Set) : Not (P /\ Q) <-> Not P \/ Not Q
notand'' {P}{Q} =
    let lem1 : Not (P /\ Q) -> Not P \/ Not Q
        lem1 h = orE (LEM P)
                 (\p -> orE (LEM Q)
                        (\q -> whenFalse (Not P \/ Not Q) (h (p,q)))
                        (\nq -> orI2 nq))
                 (\np -> orI1 np)
        lem2 : Not P \/ Not Q -> Not (P /\ Q)
        lem2 npornq pandq = orE npornq
                            (\np -> np pandq.1)
                            (\nq -> nq pandq.2)
    in (lem1, lem2)

taut5'  (P,Q,R,S:Pred X) :
          ((x:X) -> P x -> (Q x \/ (R x \/ S x))) ->
          (((x:X) -> P x -> Q x) 
           \/ (Exists (\x -> P x /\ R x) 
              \/ Exists (\x -> P x /\ S x)))
taut5' {P}{Q}{R}{S} all =
    let lem1 : Not ((x:X) -> P x -> Q x) -> Exists (\x -> P x /\ Not (Q x))
        lem1 = notall2ex'
        lem2 : Exists (\x -> P x /\ Not (Q x)) -> Exists (\x -> R x \/ S x)
        lem2 h = let x : X
                     x = h.1
                     lem21 : Q x \/ (R x \/ S x)
                     lem21 = all h.1 h.2.1
                 in (x, orE lem21
                          (\Qx -> whenFalse (R x \/ S x) (h.2.2 Qx))
                          (\RxorSx -> RxorSx))
        lem3 : Exists (\x -> P x /\ Not (Q x)) -> Exists (\x -> P x /\ R x) \/ Exists (\x -> P x /\ S x)
        lem3 h = let x : X
                     x = h.1
                     RorS : R x \/ S x
                     RorS = (lem2 h).2
                     Px : P x
                     Px = h.2.1
                 in orE RorS 
                        (\Rx -> orI1 (x, (Px, Rx))) 
                        (\Sx -> orI2 (x, (Px, Sx)))
    in orE (LEM ((x:X) -> P x -> Q x))
           (\l -> orI1 l)
           (\r -> orI2 (lem3 (lem1 r)))


taut5  (M,N,P,Q,R,S:Pred X) :
          ((x:X) -> M x -> N x -> P x -> (Q x \/ (R x \/ S x))) ->
          (((x:X) -> M x -> N x -> P x -> Q x) 
           \/ (Exists (\x -> M x /\ N x /\ P x /\ R x) 
              \/ Exists (\x -> M x /\ N x /\ P x /\ S x)))
taut5 {M}{N}{P}{Q}{R}{S} all =
    let lem1 : Not ((x:X) -> M x -> N x -> P x -> Q x) -> 
            Exists (\x -> M x /\ N x /\ P x /\ Not (Q x))
        lem1 = notall2ex''
        lem2 : Exists (\x -> M x /\ N x /\ P x /\ Not (Q x)) -> Exists (\x -> R x \/ S x)
        lem2 h = let x : X
                     x = h.1
                     lem21 : Q x \/ (R x \/ S x)
                     lem21 = all h.1 h.2.1.1.1 h.2.1.1.2 h.2.1.2
                 in (x, orE lem21
                          (\Qx -> whenFalse (R x \/ S x) (h.2.2 Qx))
                          (\RxorSx -> RxorSx))
        lem3 : Exists (\x -> M x /\ N x /\ P x /\ Not (Q x)) -> Exists (\x -> M x /\ N x /\ P x /\ R x) \/ Exists (\x -> M x /\ N x /\ P x /\ S x)
        lem3 h = let x : X
                     x = h.1
                     RorS : R x \/ S x
                     RorS = (lem2 h).2
                     Mx : M x
                     Mx = h.2.1.1.1
                     Nx : N x
                     Nx = h.2.1.1.2
                     Px : P x
                     Px = h.2.1.2
                 in orE RorS 
                        (\Rx -> orI1 (x, (((Mx, Nx), Px), Rx))) 
                        (\Sx -> orI2 (x, (((Mx, Nx), Px), Sx)))
    in orE (LEM ((x:X) -> M x -> N x -> P x -> Q x))
           (\l -> orI1 l)
           (\r -> orI2 (lem3 (lem1 r)))

eqtrans1' (a,b,c:X)(U:Pow)((<):Rel X):trich U (<) -> trans U (<) -> el a U -> el b U -> el c U -> a==b -> b<c -> a<c
eqtrans1' {a} {b} {c} {U} {(<)} tri tra aU bU cU eqab ltbc =
    orE (LEM (a<c))
           (\h -> h)
           (\notltac ->
                let lem21 : a==c \/ c<a
                    lem21 = notand' ((contrap (tri a c aU cU).1.2) notltac)
                    lem22 : a==c ->False
                    lem22 eqac = ((tri b c bU cU).2.1 
                                  (eqXtra b a c 
                                   (eqXsym a b eqab) eqac)).1 ltbc
                    lem23 : c<a -> False
                    lem23 ltca = ((tri b a bU aU).1.1 
                                  (tra b c a bU cU aU ltbc ltca)).2 
                                 (eqXsym a b eqab)
                in whenFalse (a<c) (orE lem21 lem22 lem23))

eqtrans1 (a,b,c:X)(U:Pow)((<):Rel X):GS U (<) -> el a U -> el b U -> el c U ->
                                                 a==b -> b<c -> a<c
eqtrans1 {a} {b} {c} {U} {(<)} hGS = eqtrans1' hGS.1.1.1 hGS.1.1.2

eqtrans2' (a,b,c:X)(U:Pow)((<):Rel X):trich U (<) -> trans U (<) -> 
           el a U -> el b U -> el c U -> a==b -> c<b -> c<a
eqtrans2' {a} {b} {c} {U} {(<)} tri tra aU bU cU eqab ltcb =
    orE (LEM (c<a))
           (\h -> h)
           (\notltca -> 
            let lem31 : c==a \/ a<c
                lem31 = notand' ((contrap ((tri c a cU aU).1.2)) notltca)
                lem32 : c==a -> False
                lem32 h = ((tri c b cU bU).2.1 
                           (eqXtra c a b h 
                            eqab)).1 ltcb
                lem33 : a<c -> False
                lem33 h = ((tri a b aU bU).1.1 
                           (tra a c b aU cU bU h ltcb)).2 
                          eqab
            in whenFalse (c<a) (orE lem31 lem32 lem33))

eqtrans2 (a,b,c:X)(U:Pow)((<):Rel X):GS U (<) -> el a U -> el b U -> el c U ->
                                                 a==b -> c<b -> c<a
eqtrans2 {a} {b} {c} {U} {(<)} hGS = eqtrans2' hGS.1.1.1 hGS.1.1.2

ISLem1 (U:Pow) ((<):Rel X) : GS U (<) -> (x,y:X) -> el x U -> el y U ->
                x==y -> (z:X) -> el z (IS U x (<)) -> el z (IS U y (<))
ISLem1 {U} {(<)} hGS x y xU yU eqxy z zSx = 
    let lem1 : el z U /\ z<x /\ ((w:X) -> w==z -> el w U /\ w<x)
        lem1 = (comprLem z).1 zSx
        tri : trich U (<)
        tri = hGS.1.1.1
        tra : trans U (<)
        tra = hGS.1.1.2
        lem3 : (u:X) -> el u U -> u<x -> u<y
        lem3 u uU ltux = eqtrans2 hGS yU xU uU (eqXsym x y eqxy) ltux
        lem4 : el z U /\ z<y /\ ((w:X) -> w==z -> el w U /\ w<y)
        lem4 = 
          let zU : el z U
              zU = lem1.1.1
              zlty : z<y
              zlty = lem3 z lem1.1.1 lem1.1.2
              lem2 : (w:X) -> el w U -> w==z -> z<y -> w<y
              lem2 w wU eqwz ltzy = eqtrans1 hGS wU zU yU eqwz ltzy
              extw : (w:X) -> w==z -> el w U /\ w<y
              extw w eqwz = let wU = U.2 z w (eqXsym w z eqwz) zU
                            in (wU ,lem2 w wU eqwz zlty)
          in ((zU,zlty),extw)
    in (comprLem z).2 lem4

equalEquiv : EquivRel equal
equalEquiv = 
         let isRefl : Reflexive equal 
             isRefl = \U -> (\x xU -> xU, \x xU -> xU)
             isSymm : Symmetrical equal
             isSymm = \U V eqUV -> (eqUV.2, eqUV.1)
             isTrans : Transitive equal
             isTrans = \U V W eqUV eqVW -> (\x xU -> eqVW.1 x (eqUV.1 x xU),
                                            \x xW -> eqUV.2 x (eqVW.2 x xW))
         in (isRefl, isSymm, isTrans)

ISLem2 (U1,U2:Pow) (R1,R2:Rel X) : GS U1 R1 /\ GS U2 R2 -> 
       (x,y:X) -> el x U1 -> el x U2 -> el y U1 -> el y U2 -> 
       x==y -> equal (IS U1 x R1) (IS U2 x R2) -> 
               equal (IS U1 y R1) (IS U2 y R2)
ISLem2 {U1} {U2} {R1} {R2} hGSs x y xU1 xU2 yU1 yU2 eqxy eqS1xS2x =
    let S1x = IS U1 x R1
        S1y = IS U1 y R1
        S2x = IS U2 x R2
        S2y = IS U2 y R2
        eqS1xS1y : equal S1x S1y
        eqS1xS1y = (\u uS1x -> ISLem1 hGSs.1 x y xU1 yU1 eqxy u uS1x,
                    \u uS1y -> ISLem1 hGSs.1 y x yU1 xU1 
                               (eqXsym x y eqxy) u uS1y)
        eqS2xS2y : equal S2x S2y
        eqS2xS2y = (\u uS2x -> ISLem1 hGSs.2 x y xU2 yU2 eqxy u uS2x,
                    \u uS2y -> ISLem1 hGSs.2 y x yU2 xU2 
                               (eqXsym x y eqxy) u uS2y)
        eTrans = equalEquiv.2.2
        eSymm = equalEquiv.2.1
        eqS1yS2y : equal S1y S2y
        eqS1yS2y = eTrans S1y S2x S2y
                   (eTrans S1y S1x S2x
                    (eSymm S1x S1y eqS1xS1y) eqS1xS2x)
                   eqS2xS2y
    in eqS1yS2y

ISLem3 (U1,U2:Pow) (R1,R2:Rel X) : GS U1 R1 /\ GS U2 R2 ->
       (x,y:X) -> el x U1 -> el y U1 -> 
       x==y -> equal (IS U1 x R1) U2 -> 
               equal (IS U1 y R1) U2
ISLem3 {U1} {U2} {R1} {R2} hGSs x y xU1 yU1 eqxy eqS1xU2 = 
    let S1x = IS U1 x R1
        S1y = IS U1 y R1
        eqS1xS1y : equal S1x S1y
        eqS1xS1y = (\u uS1x -> ISLem1 hGSs.1 x y xU1 yU1 eqxy u uS1x,
                    \u uS1y -> ISLem1 hGSs.1 y x yU1 xU1 
                               (eqXsym x y eqxy) u uS1y)
        eTrans = equalEquiv.2.2
        eSymm = equalEquiv.2.1
        eqS1yU2 : equal S1y U2
        eqS1yU2 = eTrans S1y S1x U2
                  (eSymm S1x S1y eqS1xS1y) eqS1xU2
    in eqS1yU2

ISLem4 (U:Pow) ((<):Rel X) : GS U (<) -> (x:X) -> el x U -> 
                                         Not (el x (IS U x (<)))
ISLem4 {U} {(<)} hGS x xU xS1x = 
    let lem1 : x<x
        lem1 = ((comprLem x).1 xS1x).1.2
        tri : trich U (<)
        tri = hGS.1.1.1
        lem2 : Not (x==x)
        lem2 = ((tri x x xU xU).1.1 lem1).2
    in lem2 (eqXref x)

ISLem5 (U:Pow) ((<):Rel X) : GS U (<) -> (x:X) -> el x U -> 
                                         el x (C (IS U x (<)))
ISLem5 {U} {(<)} hGS x xU = 
    let S : X -> Pow
        S x' = IS U x' (<)
        nxSx : Not (el x (S x))
        nxSx = ISLem4 hGS x xU
        ext : (y:X) -> y==x -> Not (el y (S x))
        ext y eqyx ySx = nxSx ((S x).2 y x eqyx ySx)
    in (comprLem {\z -> Not (el z (S x))} x).2 (nxSx,ext)


CLem0 (U,V:Pow) : equal U V -> sub (C U) (C V)
CLem0 {U} {V} eqUV x xCU = 
    let lem1 : Not (el x U)
        lem1 = ((comprLem x).1 xCU).1
        lem2 : Not (el x V)
        lem2 xV = lem1 (eqUV.2 x xV)
        ext : (w:X) -> w==x -> Not (el w V)
        ext w eqwx wV = lem2 (V.2 w x eqwx wV)
    in (comprLem x).2 (lem2,ext)

CLem1 (U,V:Pow) : equal U V -> equal (C U) (C V)
CLem1 eqUV = (CLem0 eqUV, CLem0 (eqUV.2, eqUV.1))

subLem1 (U,V:Pow) : Not (equal U V) -> sub U V -> nonempty (V-U)
subLem1 {U} {V} neqUV subUV =
    let lem1 : Not (sub U V) \/ Not (sub V U)
        lem1 = notand''.1 neqUV
        lem2 : Not (sub U V) -> nonempty (V-U)
        lem2 h = whenFalse (nonempty (V-U)) (h subUV)
        lem3 : Not (sub V U) -> nonempty (V-U)
        lem3 h = let witness : Exists (\x -> el x V /\ Not (el x U))
                     witness = notall2ex' h
                     x = witness.1
                     xV = witness.2.1
                     nxU = witness.2.2
                     ext : (w:X) -> w==x -> el w V /\ Not (el w U)
                     ext w eqwx = (V.2 x w (eqXsym w x eqwx) xV
                                  ,(contrap (U.2 w x eqwx)) nxU)
                     xVminusU : el x (V-U)
                     xVminusU = (comprLem x).2 ((xV,nxU),ext)
                 in (x, xVminusU)
    in orE lem1 lem2 lem3

Lemma (M1,M2:Pow) (R1,R2:Rel X) : GS M1 R1 /\ GS M2 R2 ->
    (x:X) -> el x M1 ->
    (el x M2 /\ equal (IS M1 x R1) (IS M2 x R2)) \/ 
    (Exists (\x1 -> el x1 (IS M1 x R1) /\ equal (IS M1 x1 R1) M2) \/ 
     equal (IS M1 x R1) M2)
Lemma {M1}{M2}{R1}{R2} hGSs z hM1z =
 let S1 : X -> Pow
     S1 = \x -> IS M1 x R1
     S2 : X -> Pow
     S2 = \x -> IS M2 x R2
     P : Pred X
     P x= el x M1 ->
          (el x M2 /\ equal (S1 x) (S2 x)) \/ 
          (Exists (\x1 -> el x1 (S1 x) /\ equal (S1 x1) M2) \/ 
           equal (S1 x) M2)
     ExtP : (x,y:X) -> x==y -> P x -> P y
     ExtP x y eqxy Px yM1 =
         let xM1 : el x M1
             xM1 = M1.2 y x (eqXsym x y eqxy) yM1
             lem1 : el x M2 /\ equal (S1 x) (S2 x) ->
                    el y M2 /\ equal (S1 y) (S2 y)
             lem1 h = let yM2 = M2.2 x y eqxy h.1
                      in (yM2, ISLem2 hGSs x y xM1 h.1 yM1 yM2 eqxy h.2)
             lem2 : Exists (\x1 -> el x1 (S1 x) /\ equal (S1 x1) M2) ->
                    Exists (\x1 -> el x1 (S1 y) /\ equal (S1 x1) M2)
             lem2 h = let x1 = h.1
                          x1S1x = h.2.1
                          hiso = h.2.2
                          x1S1y = ISLem1 hGSs.1 x y xM1 yM1 eqxy x1 x1S1x
                      in (x1, (x1S1y, hiso))
             lem3 : equal (S1 x) M2 -> equal (S1 y) M2
             lem3 h = ISLem3 hGSs x y xM1 yM1 eqxy h
         in orE (Px xM1) 
            (\h -> orI1 (lem1 h)) 
            (\h -> orI2 (orE h (\h -> orI1 (lem2 h)) (\h -> orI2 (lem3 h))))
     WFM1 : wellFounded M1 R1 (P,ExtP)
     WFM1 = WO2WF M1 R1 (P,ExtP) hGSs.1.1
     (<) = R1
     induct : (x:X) -> el x M1 -> ((y:X) -> el y M1 -> y<x -> P y) -> P x
     induct x xM1 IH = 
       let split = taut5 IH
           case1 : ((y:X) -> el y M1 -> y<x -> el y M1 ->
                             el y M2 /\ equal (S1 y) (S2 y)) ->
                   P x
           case1 IH' = 
               let subS1xM2 : sub (S1 x) M2
                   subS1xM2 y yS1x =
                       let yM1 : el y M1
                           yM1 = ((comprLem y).1 yS1x).1.1
                           ltyx : y<x
                           ltyx = ((comprLem y).1 yS1x).1.2
                           yM2 : el y M2
                           yM2 = (IH' y yM1 ltyx yM1).1
                       in yM2
                   case1a : equal (S1 x) M2 -> equal (S1 x) M2
                   case1a eqS1xM2 = eqS1xM2
                   case1b : Not (equal (S1 x) M2) -> el x M2 /\ equal (S1 x) (S2 x)
                   case1b neqS1xM2 = 
                       let V : Pow
                           V = M2 - (S1 x)
                           neV : nonempty V
                           neV = subLem1 neqS1xM2 subS1xM2
                           V' : Pow'
                           V' = (V,neV)
                           subVM2 : sub V M2
                           subVM2 y yV = ((comprLem y).1 yV).1.1
                           t : X
                           t = minimal M2 R2 V' hGSs.2.1 subVM2
                           tV : el t V
                           tV = minimalLem M2 R2 V' hGSs.2.1 subVM2
                           tM2 : el t M2
                           tM2 = subVM2 t tV
                           subS1xS2t : sub (S1 x) (S2 t)
                           subS1xS2t a aS1x = 
                               let (<) = R2
                                   tri : trich M2 (<)
                                   tri = hGSs.2.1.1.1
                                   aM2 : el a M2
                                   aM2 = subS1xM2 a aS1x
                                   lem1 : a<t -> el a (S2 t)
                                   lem1 ltat = 
                                     let ext : (w:X) -> w==a -> el w M2 /\ w<t
                                         ext w eqwa = 
                                           let wM2 = M2.2 a w (eqXsym w a eqwa) aM2
                                           in (wM2, eqtrans1 hGSs.2 wM2 aM2 tM2  eqwa ltat)
                                     in (comprLem a).2 ((aM2, ltat), ext)
                                   lem2 : Not (a<t) -> el a (S2 t)
                                   lem2 nltat = 
                                     let lem21 : t<a \/ a==t
                                         lem21 = notand.1 ((contrap ((tri a t aM2 tM2).1.2)) nltat)
                                         ntS1x : Not (el t (S1 x))
                                         ntS1x = ((comprLem t).1 tV).1.2
                                         lem22 : t<a -> False
                                         lem22 ltta = 
                                           let ltax : R1 a x
                                               ltax = ((comprLem a).1 aS1x).1.2
                                               aM1 : el a M1
                                               aM1 = ((comprLem a).1 aS1x).1.1
                                               eqS1aS2a : equal (S1 a) (S2 a)
                                               eqS1aS2a = (IH' a aM1 ltax aM1).2
                                               tS2a : el t (S2 a)
                                               tS2a =
                                                 let ext : (w:X) -> w==t -> el w M2 /\ w<a
                                                     ext w eqwt = 
                                                       let wM2 = M2.2 t w (eqXsym w t eqwt) tM2
                                                       in (wM2, eqtrans1 hGSs.2 wM2 tM2 aM2 eqwt ltta)
                                                 in (comprLem t).2 ((tM2,ltta),ext)
                                               tS1a : el t (S1 a)
                                               tS1a = eqS1aS2a.2 t tS2a
                                               tM1 : el t M1
                                               tM1 = ((comprLem t).1 tS1a).1.1
                                               R1ta : R1 t a
                                               R1ta = ((comprLem t).1 tS1a).1.2
                                               tra : trans M1 R1
                                               tra = hGSs.1.1.1.2
                                               R1tx : R1 t x
                                               R1tx = tra t a x tM1 aM1 xM1 R1ta ltax
                                               ext : (w:X) -> w==t -> el w M1 /\ R1 w x
                                               ext w eqwt = 
                                                 let wM1 : el w M1
                                                     wM1 = M1.2 t w (eqXsym w t eqwt) tM1
                                                     R1wx : R1 w x
                                                     R1wx = eqtrans1 hGSs.1 wM1 tM1 xM1 eqwt R1tx
                                                 in (wM1, R1wx)
                                               tS1x : el t (S1 x)
                                               tS1x = (comprLem t).2 ((tM1,R1tx),ext)
                                           in ntS1x tS1x
                                         lem23 : a==t -> False
                                         lem23 eqat =
                                           let tS1x : el t (S1 x)
                                               tS1x = (S1 x).2 a t eqat aS1x
                                           in ntS1x tS1x
                                     in whenFalse (el a (S2 t)) (orE lem21 lem22 lem23)
                               in orE (LEM (a<t)) lem1 lem2
                           subS2tS1x : sub (S2 t) (S1 x)
                           subS2tS1x b bS2t = 
                               let (<) = R2
                                   ltbt : b < t
                                   ltbt = ((comprLem b).1 bS2t).1.2
                                   lem1 : Not (el b (S1 x)) -> False
                                   lem1 nelbS1x = 
                                       let bM2 : el b M2
                                           bM2 = ((comprLem b).1 bS2t).1.1
                                           ext : ((w:X) -> w==b -> el w M2 /\ Not (el w (S1 x)))
                                           ext w eqwb = 
                                               (M2.2 b w (eqXsym w b eqwb) bM2, 
                                                (contrap ((S1 x).2 w b eqwb)) nelbS1x)
                                           bV : el b V
                                           bV = (comprLem b).2 ((bM2,nelbS1x),ext)
                                           nltbt : Not (b < t)
                                           nltbt = minimalLem2 M2 R2 V' 
                                                   hGSs.2.1 subVM2 b bV
                                       in nltbt ltbt
                               in orE (LEM (el b (S1 x))) 
                                      (\h -> h) 
                                      (\h -> whenFalse (el b (S1 x)) (lem1 h))
                           eqS1xS2t : equal (S1 x) (S2 t)
                           eqS1xS2t = (subS1xS2t, subS2tS1x)
                           xCS1x : el x (C (S1 x))
                           xCS1x = ISLem5 hGSs.1 x xM1
                           tCS2t : el t (C (S2 t))
                           tCS2t = ISLem5 hGSs.2 t tM2
                           neCS1x : nonempty (C (S1 x))
                           neCS1x = (x, xCS1x)
                           neCS2t : nonempty (C (S2 t))
                           neCS2t = (t, tCS2t)
                           g1 = gamma (C (S1 x), neCS1x)
                           g2 = gamma (C (S2 t), neCS2t)
                           eqxg1 : x == g1
                           eqxg1 = hGSs.1.2 x xM1 neCS1x
                           eqtg2 : t == g2
                           eqtg2 = hGSs.2.2 t tM2 neCS2t
                           eqCS1xCS2t : equal (C (S1 x)) (C (S2 t))
                           eqCS1xCS2t = CLem1 eqS1xS2t
                           eqg1g2 : g1 == g2
                           eqg1g2 = AC2gamma.2.2 (C (S1 x), neCS1x) (C (S2 t), neCS2t) eqCS1xCS2t
                           eqxg2 : x == g2
                           eqxg2 = eqXtra x g1 g2 eqxg1 eqg1g2
                           eqxt : x==t
                           eqxt = eqXtra x g2 t eqxg2 (eqXsym t g2 eqtg2)
                           xM2 : el x M2
                           xM2 = M2.2 t x (eqXsym x t eqxt) tM2
                           eqS2tS2x : equal (S2 t) (S2 x)
                           eqS2tS2x = (ISLem1 hGSs.2 t x tM2 xM2 (eqXsym x t eqxt)
                                      ,ISLem1 hGSs.2 x t xM2 tM2 eqxt)
                           equalTrans = equalEquiv.2.2
                           eqS1xS2x : equal (S1 x) (S2 x)
                           eqS1xS2x = equalTrans (S1 x) (S2 t) (S2 x) eqS1xS2t eqS2tS2x
                       in (xM2, eqS1xS2x)
               in \xM1 -> orE (LEM (equal (S1 x) M2)) (\h -> orI2 (orI2 (case1a h))) (\h -> orI1 (case1b h))
           case2 : Exists (\y -> el y M1 /\ y<x /\ el y M1 /\ -- el y M2
                                 Exists (\x1 -> el x1 (S1 y) /\ 
                                                equal (S1 x1) M2)) ->
                   P x
           case2 h =
               let y = h.1
                   ltyx : y<x
                   ltyx = h.2.1.1.2
                   yM1 : el y M1
                   yM1 = h.2.1.1.1
                   x1 = h.2.2.1
                   x1S1y : el x1 (S1 y)
                   x1S1y = h.2.2.2.1
                   tra : trans M1 (<)
                   tra = hGSs.1.1.1.2
                   lem1 : el x1 M1 /\ x1<y /\ ((w:X)-> w==x1 -> el w M1 /\ w<y)
                   lem1 = (comprLem x1).1 x1S1y
                   x1M1 : el x1 M1
                   x1M1 = lem1.1.1
                   ltx1y : x1<y
                   ltx1y = lem1.1.2
                   lem2 : el x1 M1 /\ x1<x /\ ((w:X)-> w==x1 -> el w M1 /\ w<x)
                   lem2 = 
                       let ltx1x : x1<x
                           ltx1x = tra x1 y x x1M1 yM1 xM1 ltx1y ltyx
                           ext : (w:X)-> w==x1 -> el w M1 /\ w<x
                           ext w eqwx1 = 
                            let wM1 : el w M1
                                wM1 = M1.2 x1 w (eqXsym w x1 eqwx1) x1M1
                            in (wM1, eqtrans1 hGSs.1 wM1 x1M1 xM1 eqwx1 ltx1x)
                       in ((x1M1, ltx1x), ext)
                   x1S1x : el x1 (S1 x)
                   x1S1x = (comprLem x1).2 lem2
                   hiso = h.2.2.2.2
               in \void -> orI2 (orI1 (x1, (x1S1x, hiso)))
           case3 : Exists (\y -> el y M1 /\ y<x /\ el y M1 /\ 
                           equal (S1 y) M2) -> 
                   P x
           case3 h =
               let y = h.1
                   ltyx : y<x
                   ltyx = h.2.1.1.2
                   yM1 : el y M1
                   yM1 = h.2.1.1.1
                   ext : (w:X) -> w==y -> el w M1 /\ w<x
                   ext w eqwy = 
                       let wM1 = M1.2 y w (eqXsym w y eqwy) yM1
                           ltwx = eqtrans1 hGSs.1 wM1 yM1 xM1 eqwy ltyx
                       in (wM1, ltwx)
                   yS1x : el y (S1 x)
                   yS1x = (comprLem y).2 ((yM1,ltyx),ext)
               in \void -> orI2 (orI1 (y, (yS1x, h.2.2)))
       in orE split 
          (\h -> case1 h)
          (\h -> orE h (case2) (case3))
 in WFM1 induct z hM1z hM1z

settaut1' (A,B:Pow)(x:X) : el x A -> Not (el x B) -> el x (A-B)
settaut1' {A}{B}{x} xA nxB = 
    let ext : (y:X) -> y==x -> el y A /\ Not (el y B)
        ext y eqyx = 
            (A.2 x y (eqXsym y x eqyx) xA, 
            \yB -> nxB (B.2 y x eqyx yB))
    in (comprLem x).2 ((xA, nxB), ext)

settaut1 (A,B:Pow) : sub A B -> equal O (B-A) -> equal A B
settaut1 {A}{B} subAB eqOBmA = 
    let subBA : sub B A
        subBA x xB = 
            let alt1 : el x A -> el x A
                alt1 xA = xA
                alt2 : Not (el x A) -> el x A
                alt2 nxA =
                    let xBmA : el x (B-A)
                        xBmA = settaut1' xB nxA
                        xO : el x O
                        xO = eqOBmA.2 x xBmA
                    in whenFalse (el x A) ((comprLem x).1 xO).1
            in orE (LEM (el x A)) alt1 alt2
    in (subAB, subBA)

settaut2' (A:Pow) : sub O A
settaut2' {A} x xO = whenFalse (el x A) ((comprLem x).1 xO).1

settaut2 (A:Pow) : Not (equal A O) -> nonempty A
settaut2 {A} neqAO = 
    let lem1 : Not (sub A O) \/ Not (sub O A)
        lem1 = notand''.1 neqAO
        lem2 : Not (sub O A) -> nonempty A
        lem2 h = whenFalse (nonempty A) (h settaut2')
        lem3 : Not (sub A O) -> nonempty A
        lem3 h = ((notall2ex' h).1, (notall2ex' h).2.1)
    in orE lem1 lem3 lem2

settaut3 (x:X)(A,B:Pow) : Not (el x (A-B)) -> el x A -> el x B
settaut3 {x}{A}{B} nxAmB xA = 
       orE (LEM (el x B))
           (\xB -> xB)
           (\nxB -> whenFalse (el x B) (nxAmB (settaut1' xA nxB)))

settaut8 (A,B:Pow) : sub (A-B) A
settaut8 {A}{B} x xAmB = ((comprLem x).1 xAmB).1.1

settaut4' (A,B,D:Pow) : sub (A-B) (D-B) -> sub B D -> sub A D
settaut4' {A}{B}{D} subAmBDmB subBD x xA =
    let lem1 : Not (el x B) ->  el x D
        lem1 nxB = settaut8 x (subAmBDmB x (settaut1' xA nxB))
        lem2 : el x B -> el x D
        lem2 xB = subBD x xB
    in orE (LEM (el x B)) lem2 lem1

settaut4 (A,B,D:Pow) : equal (A-B) (D-B) -> sub B A -> sub B D -> equal A D
settaut4 {A}{B}{D} h h1 h2 = (settaut4' h.1 h2, settaut4' h.2 h1)

settaut5 (A,B,D:Pow) : sub A D -> sub (A-B) (D-B)
settaut5 {A}{B}{D} subAD x xAmB = 
    let xA : el x A
        xA = ((comprLem x).1 xAmB).1.1
        nxB : Not (el x B)
        nxB = ((comprLem x).1 xAmB).1.2
    in settaut1' (subAD x xA) nxB

settaut6' (A,B:Pow)(x:X) : el x A -> el x (A+B)
settaut6' {A}{B}{x} xA = 
    let ext : (y:X) -> y==x -> el y A \/ el y B
        ext y eqyx = orI1 (A.2 x y (eqXsym y x eqyx) xA)
    in (comprLem x).2 (orI1 xA, ext)

settaut6'' (A,B:Pow)(x:X) : el x B -> el x (A+B)
settaut6'' {A}{B}{x} xB = 
    let ext : (y:X) -> y==x -> el y A \/ el y B
        ext y eqyx = orI2 (B.2 x y (eqXsym y x eqyx) xB)
    in (comprLem x).2 (orI2 xB, ext)

settaut6 (A,B:Pow) : sub B A -> equal A (A-B+B)
settaut6 {A}{B} subBA =
    let sub1 : sub A (A-B+B)
        sub1 x xA = 
            let lem1 : el x B -> el x (A-B+B)
                lem1 xB = settaut6'' xB
                lem2 : Not (el x B) -> el x (A-B+B)
                lem2 nxB = settaut6' (settaut1' xA nxB)
            in orE (LEM (el x B)) lem1 lem2
        sub2 : sub (A-B+B) A
        sub2 x xAmBpB = 
            orE ((comprLem x).1 xAmBpB).1
                    (\xAmB -> settaut8 x xAmB)
                    (\xB -> subBA x xB)
    in (sub1, sub2)

settaut9 (A,B:Pow) : equal A O -> sub B A -> equal B O
settaut9 {A}{B} eqAO subBA = 
    let lem1 : sub B O
        lem1 x xB = eqAO.1 x (subBA x xB)
    in (lem1, settaut2')

settaut11' : (s,t:X) -> s==t -> el s (Singl t)
settaut11' s t eqst = 
    let ext : (y:X) -> y==s -> y==t
        ext y eqys = eqXtra y s t eqys eqst
    in (comprLem s).2 (eqst, ext)

settaut11 : (t:X) -> el t (Singl t)
settaut11 t = settaut11' t t (eqXref t)

settaut10 (A,B:Pow) (t:X): equal (A-(Singl t)) O -> sub (Singl t) A ->
                                               equal A (Singl t)
settaut10 {A}{B}{t} h1 h2 = 
    let symm : Symmetrical equal
        symm = equalEquiv.2.1
    in symm (Singl t) A (settaut1 h2 (symm (A-(Singl t)) O h1))

Step5 (M1,M2:Pow) (R1,R2:Rel X) : GS M1 R1 /\ GS M2 R2 ->
    equal M1 M2 \/ 
    (Exists (\x1 -> el x1 M1 /\ equal (IS M1 x1 R1) M2) \/
     Exists (\x2 -> el x2 M2 /\ equal (IS M2 x2 R2) M1))
Step5 {M1} {M2} {R1} {R2} hGSs =
 let S1 : X -> Pow
     S1 x = IS M1 x R1
     S2 : X -> Pow
     S2 x = IS M2 x R2
     Goal = equal M1 M2 \/ 
            (Exists (\x1 -> el x1 M1 /\ equal (S1 x1) M2) \/
             Exists (\x2 -> el x2 M2 /\ equal (S2 x2) M1))
     case1 : ((x:X) -> el x M1 ->
              (el x M2 /\ equal (S1 x) (S2 x))) 
             -> Goal
     case1 hyp = 
      let subM1M2 : sub M1 M2
          subM1M2 x xM1 = (hyp x xM1).1
          S = compr (\y -> Exists (\z -> el z M1 /\ el y (S1 z)))
          T1 = M1 - S
          T2 = M2 - S
          subSM1 : sub S M1
          subSM1 y yS = ((comprLem y).1 ((comprLem y).1 yS).1.2.2).1.1
          subSM2 : sub S M2
          subSM2 y yS = subM1M2 y (subSM1 y yS)
          subT1M1 : sub T1 M1
          subT1M1 a aT1 = ((comprLem a).1 aT1).1.1
          subT2M2 : sub T2 M2
          subT2M2 a aT2 = ((comprLem a).1 aT2).1.1
          WOM2 : wellOrdered M2 R2
          WOM2 = hGSs.2.1
          lema : (neqT2O : Not (equal T2 O)) -> equal (S2 (minimal M2 R2 (T2,settaut2 neqT2O) WOM2 subT2M2)) S
          lema neqT2O =
                let neT2 : nonempty T2
                    neT2 = settaut2 neqT2O
                    t : X
                    t = minimal M2 R2 (T2,neT2) WOM2 subT2M2
                    tT2 : el t T2
                    tT2 = minimalLem M2 R2 (T2,neT2) WOM2 subT2M2
                    tM2 : el t M2
                    tM2 = subT2M2 t tT2
                    subS2tM2 : sub (S2 t) M2
                    subS2tM2 a aS2t = ((comprLem a).1 aS2t).1.1
                    subSS2t : sub S (S2 t)
                    subSS2t a aS = 
                        let naT2 : Not (el a T2)
                            naT2 aT2 = let naS : Not (el a S)
                                           naS = ((comprLem a).1 aT2).1.2
                                       in naS aS
                            aM1 : el a M1
                            aM1 = subSM1 a aS
                            aM2 : el a M2
                            aM2 = subM1M2 a aM1
                            lem1 : R2 t a -> False
                            lem1 R2ta = 
                             let eqS1aS2a : equal (S1 a) (S2 a)
                                 eqS1aS2a = (hyp a aM1).2
                                 tS2a : el t (S2 a)
                                 tS2a = let ext : (w:X) -> w==t -> el w M2 /\ R2 w a
                                            ext w eqwt = let wM2 = M2.2 t w (eqXsym w t eqwt) tM2
                                                         in (wM2, eqtrans1 hGSs.2 wM2 tM2 aM2 eqwt R2ta)
                                        in (comprLem t).2 ((tM2,R2ta), ext)
                                 tS1a : el t (S1 a)
                                 tS1a = eqS1aS2a.2 t tS2a
                                 tS : el t S
                                 tS = let lem1 : Exists (\z -> el z M1 /\ el t (S1 z))
                                          lem1 = (a, (aM1, tS1a))
                                          lem2 : (w:X) -> w==t -> Exists (\z -> el z M1 /\ el w (S1 z))
                                          lem2 w eqwt = (a, (aM1, (S1 a).2 t w (eqXsym w t eqwt) tS1a))
                                      in (comprLem t).2 (lem1, lem2)
                                 ntS : Not (el t S)
                                 ntS = ((comprLem t).1 tT2).1.2
                             in ntS tS
                            lem2 : a==t -> False
                            lem2 h = naT2 (T2.2 t a (eqXsym a t h) tT2)
                            tri : trich M2 R2
                            tri = hGSs.2.1.1.1
                            lem3 : Not (R2 a t) -> R2 t a \/ a==t
                            lem3 h = notand.1 ((contrap ((tri a t aM2 tM2).1.2)) h)
                            R2at : R2 a t
                            R2at = orE (LEM (R2 a t))
                                   (\h -> h)
                                   (\h -> whenFalse (R2 a t) (orE (lem3 h) lem1 lem2))
                            ext : (w:X) -> w==a -> el w M2 /\ R2 w t
                            ext w eqwa = 
                                let wM2 = M2.2 a w (eqXsym w a eqwa) aM2
                                in (wM2, eqtrans1 hGSs.2 wM2 aM2 tM2 eqwa R2at)
                            aS2t : el a (S2 t)
                            aS2t = (comprLem a).2 ((aM2, R2at), ext)
                        in aS2t
                    subS2tS : sub (S2 t) S
                    subS2tS a aS2t = 
                     let lem1 : (el a T2) -> Not (R2 a t)
                         lem1 h = minimalLem2 M2 R2 (T2,neT2) WOM2 subT2M2 a h
                         R2at : R2 a t
                         R2at = ((comprLem a).1 aS2t).1.2
                         naT2 : Not (el a T2)
                         naT2 = orE (LEM (el a T2))
                                (\h -> whenFalse (Not (el a T2))((lem1 h) R2at))
                                (\h -> h)
                         aM2 : el a M2
                         aM2 = subS2tM2 a aS2t
                     in settaut3 naT2 aM2
                in (subS2tS, subSS2t)
          equalSymm = equalEquiv.2.1
          equalTrans = equalEquiv.2.2
          casea : equal T1 O -> Goal
          casea eqT1O = 
           let eqSM1 : equal S M1
               eqSM1 = settaut1 subSM1 (equalSymm T1 O eqT1O)
               eqM1S : equal M1 S
               eqM1S = equalSymm S M1 eqSM1
               caseai : equal T2 O -> Goal
               caseai eqT2O = 
                let eqSM2 : equal S M2
                    eqSM2 = settaut1 subSM2 (equalSymm T2 O eqT2O)
                    eqM1M2 : equal M1 M2
                    eqM1M2 = equalTrans M1 S M2 eqM1S eqSM2
                in orI1 eqM1M2
               caseaii : Not (equal T2 O) -> Goal
               caseaii neqT2O = 
                let neT2 : nonempty T2
                    neT2 = settaut2 neqT2O
                    t : X
                    t = minimal M2 R2 (T2,neT2) WOM2 subT2M2
                    tT2 : el t T2
                    tT2 = minimalLem M2 R2 (T2,neT2) WOM2 subT2M2
                    tM2 : el t M2
                    tM2 = subT2M2 t tT2
                    eqS2tM1 : equal (S2 t) M1
                    eqS2tM1 = equalTrans (S2 t) S M1
                              (lema neqT2O)
                              (equalSymm M1 S eqM1S)
                in orI2 (orI2 (t, (tM2,eqS2tM1)))
           in orE (LEM (equal T2 O)) caseai caseaii
          caseb : Not (equal T1 O) -> Goal
          caseb neqT1O =
           let tri : trich M1 R1
               tri = hGSs.1.1.1.1
               singlT1 : (t1,t2:X) -> el t1 T1 -> el t2 T1 -> t1==t2
               singlT1 t1 t2 t1T1 t2T1 = 
                let (<) = R1
                    alts : Not (t1==t2) -> t1<t2 \/ t2<t1
                    alts h = notand.1 ((contrap (tri t1 t2 (subT1M1 t1 t1T1) (subT1M1 t2 t2T1)).2.2) h)
                    t1M1 : el t1 M1
                    t1M1 = subT1M1 t1 t1T1
                    t2M1 : el t2 M1
                    t2M1 = subT1M1 t2 t2T1
                    alt1 : t1<t2 -> False
                    alt1 ltt1t2 = 
                     let ext : (w:X) -> w==t1 -> el w M1 /\ w<t2
                         ext w eqwt1 = 
                             let wM1 = M1.2 t1 w (eqXsym w t1 eqwt1) t1M1
                             in (wM1, eqtrans1 hGSs.1 wM1 t1M1 t2M1 eqwt1 ltt1t2)
                         t1S1t2 : el t1 (S1 t2)
                         t1S1t2 = (comprLem t1).2 ((t1M1, ltt1t2), ext)
                         ext' : (w:X) -> w==t1 -> Exists (\z -> el z M1 /\ el w (S1 z))
                         ext' w eqwt1 = (t2, (t2M1, (S1 t2).2 t1 w (eqXsym w t1 eqwt1) t1S1t2))
                         t1S : el t1 S
                         t1S = (comprLem t1).2 ((t2, (t2M1, t1S1t2)), ext')
                         nt1S : Not (el t1 S)
                         nt1S = ((comprLem t1).1 t1T1).1.2
                     in nt1S t1S
                    alt2 : t2<t1 -> False
                    alt2 ltt2t1 = 
                     let ext : (w:X) -> w==t2 -> el w M1 /\ w<t1
                         ext w eqwt2 =
                             let wM1 = M1.2 t2 w (eqXsym w t2 eqwt2) t2M1
                             in (wM1, eqtrans1 hGSs.1 wM1 t2M1 t1M1 eqwt2 ltt2t1)
                         t2S1t1 : el t2 (S1 t1)
                         t2S1t1 = (comprLem t2).2 ((t2M1, ltt2t1), ext)
                         ext' : (w:X) -> w==t2 -> Exists (\z -> el z M1 /\ el w (S1 z))
                         ext' w eqwt2 = (t1, (t1M1, (S1 t1).2 t2 w (eqXsym w t2 eqwt2) t2S1t1))
                         t2S : el t2 S
                         t2S = (comprLem t2).2 ((t1, (t1M1, t2S1t1)), ext')
                         nt2S : Not (el t2 S)
                         nt2S = ((comprLem t2).1 t2T1).1.2
                     in nt2S t2S
                in orE (LEM (t1==t2)) (\h -> h) (\h -> whenFalse (t1==t2) (orE (alts h) alt1 alt2))
               neT1 : nonempty T1
               neT1 = settaut2 neqT1O
               t : X
               t = neT1.1
               tT1 : el t T1
               tT1 = neT1.2
               eqT1Singlt : equal T1 (Singl t)
               eqT1Singlt = 
                let lem1 : sub T1 (Singl t)
                    lem1 x xT1 = 
                        let eqxt = singlT1 x t xT1 tT1
                            ext : (w:X) -> w==x -> w==t
                            ext w eqwx = eqXtra w x t eqwx eqxt
                        in (comprLem x).2 (eqxt, ext)
                    lem2 : sub (Singl t) T1
                    lem2 x xSinglt = 
                        let eqxt = ((comprLem x).1 xSinglt).1
                            eqtx = eqXsym x t eqxt
                        in T1.2 t x eqtx tT1
                in (lem1, lem2)
               casebi : equal T2 (Singl t) -> Goal
               casebi h = let eqT1T2 : equal T1 T2
                              eqT1T2 = equalTrans T1 (Singl t) T2 eqT1Singlt (equalSymm T2 (Singl t) h)
                              eqM1M2 : equal M1 M2
                              eqM1M2 = settaut4 eqT1T2 subSM1 subSM2
                          in orI1 eqM1M2
               casebii : Not (equal T2 (Singl t)) -> Goal
               casebii neqT2Singlt =
                let subT1T2 : sub T1 T2
                    subT1T2 = settaut5 subM1M2
                    neqT2O : Not (equal T2 O)
                    neqT2O eqT2O = neqT1O (settaut9 eqT2O subT1T2)
                    neT2 : nonempty T2
                    neT2 = settaut2 neqT2O
                    T2' : Pow
                    T2' = T2 - (Singl t)
                    subT2'T2 : sub T2' T2
                    subT2'T2 = settaut8
                    subT2'M2 : sub T2' M2
                    subT2'M2 x xT2' = subT2M2 x (subT2'T2 x xT2')
                    subSingltT2 : sub (Singl t) T2
                    subSingltT2 x xSinglt = subT1T2 x (eqT1Singlt.2 x xSinglt)
                    neT2' : nonempty T2'
                    neT2' = orE (LEM (nonempty T2'))
                            (\ne -> ne)
                            (\nne -> whenFalse (nonempty T2') (neqT2Singlt ((settaut10 (notnot ((contrap settaut2) nne)) subSingltT2))))
                    t' : X
                    t' = minimal M2 R2 (T2',neT2') WOM2 subT2'M2
                    t'T2' : el t' T2'
                    t'T2' = (minimalLem M2 R2 (T2',neT2') WOM2 subT2'M2)
                    t'M2 : el t' M2
                    t'M2 = subT2'M2 t' t'T2'
                    t'T2 : el t' T2
                    t'T2 = subT2'T2 t' t'T2'
                    tri : trich M2 R2
                    tri = hGSs.2.1.1.1
                    tM2 : el t M2
                    tM2 = subT2M2 t (subT1T2 t tT1)
                    nR2t't : R2 t' t -> False
                    nR2t't R2t't =
                        let tM1 : el t M1
                            tM1 = (subT1M1 t tT1)
                            eqS1tS2t : equal (S1 t) (S2 t)
                            eqS1tS2t = (hyp t tM1).2
                            t'S2t : el t' (S2 t)
                            t'S2t = 
                                let ext : (y:X) -> y==t' -> el y M2 /\ R2 y t
                                    ext y eqyt' = 
                                        let yM2 = M2.2 t' y (eqXsym y t' eqyt') t'M2
                                            R2yt = eqtrans1 hGSs.2 yM2 t'M2 tM2 eqyt' R2t't
                                        in (yM2, R2yt)
                                in (comprLem t').2 ((t'M2, R2t't), ext)
                            t'S1t : el t' (S1 t)
                            t'S1t = eqS1tS2t.2 t' t'S2t
                            t'S : el t' S
                            t'S = 
                              let lem1 : Exists (\z -> el z M1 /\ el t' (S1 z))
                                  lem1 = (t, (tM1, t'S1t))
                                  ext : (y:X) -> y==t' -> Exists (\z -> el z M1 /\ el y (S1 z))
                                  ext y eqyt' = (t, (tM1, (S1 t).2 t' y (eqXsym y t' eqyt') t'S1t))
                              in (comprLem t').2 (lem1, ext)
                            nt'S : Not (el t' S)
                            nt'S = ((comprLem t').1 t'T2).1.2
                        in nt'S t'S
                    neqtt' : Not (t==t')
                    neqtt' eqtt' = 
                        let tT2' : el t T2'
                            tT2' = T2'.2 t' t (eqXsym t t' eqtt') t'T2'
                            ntT2' : Not (el t T2')
                            ntT2' h = 
                                let ntSinglt : Not (el t (Singl t))
                                    ntSinglt = ((comprLem t).1 h).1.2
                                    tSinglt : el t (Singl t)
                                    tSinglt = settaut11 t
                                in ntSinglt tSinglt
                        in ntT2' tT2'
                    R2tt' : R2 t t'
                    R2tt' = orE (LEM (R2 t t'))
                            (\R2tt' -> R2tt')
                            (\nR2tt' -> 
                             whenFalse (R2 t t') 
                             (orE (notand.1 
                                   ((contrap (tri t t' tM2 t'M2).1.2) nR2tt'))
                              nR2t't neqtt'))
                    t'' : X
                    t'' = minimal M2 R2 (T2,neT2) WOM2 subT2M2
                    t''T2 : el t'' T2
                    t''T2 = minimalLem M2 R2 (T2,neT2) WOM2 subT2M2
                    t''M2 : el t'' M2
                    t''M2 = subT2M2 t'' t''T2
                    eqS2t''S : equal (S2 t'') S
                    eqS2t''S = lema neqT2O
                    tT2 : el t T2
                    tT2 = subT1T2 t tT1
                    tM2 : el t M2
                    tM2 = subT2M2 t tT2
                    tra : trans M2 R2
                    tra = hGSs.2.1.1.2
                    nR2tt'' : Not (R2 t t'')
                    nR2tt'' = minimalLem2 M2 R2 (T2,neT2) WOM2 subT2M2 t tT2
                    nR2t''t : Not (R2 t'' t)
                    nR2t''t R2t''t = 
                        let R2t''t' : R2 t'' t'
                            R2t''t' = tra t'' t t' t''M2 tM2 t'M2 R2t''t R2tt'
                            neqt''t : Not (t'' == t)
                            neqt''t = ((tri t'' t t''M2 tM2).1.1 R2t''t).2
                            nt''Singlt : Not (el t'' (Singl t))
                            nt''Singlt t''Singlt = 
                                neqt''t (((comprLem t'').1 t''Singlt).1)
                            t''T2' : el t'' T2'
                            t''T2' = 
                                let ext : (y:X) -> y==t'' -> el y T2 /\ Not (el y (Singl t))
                                    ext y eqyt'' = 
                                     (T2.2 t'' y (eqXsym y t'' eqyt'') t''T2,
                                     \ySinglt -> nt''Singlt ((Singl t).2 y t'' eqyt'' ySinglt))
                                in (comprLem t'').2 ((t''T2, nt''Singlt), ext)
                            nR2t''t' : Not (R2 t'' t')
                            nR2t''t' = minimalLem2 M2 R2 (T2',neT2') WOM2 subT2'M2 t'' t''T2'
                        in nR2t''t' R2t''t'
                    eqtt'' : t == t''
                    eqtt'' = orE (LEM (t==t''))
                             (\eqtt'' -> eqtt'')
                             (\neqtt'' -> 
                                  whenFalse (t==t'') 
                                  (orE (notand.1 ((contrap (tri t t'' tM2 t''M2).2.2) neqtt'')) nR2tt'' nR2t''t))
                    eqS2tS2t'' : equal (S2 t) (S2 t'')
                    eqS2tS2t'' = 
                        (ISLem1 hGSs.2 t t'' tM2 t''M2 eqtt'',
                         ISLem1 hGSs.2 t'' t t''M2 tM2 (eqXsym t t'' eqtt''))
                    eqS2tS : equal (S2 t) S
                    eqS2tS = equalEquiv.2.2 (S2 t) (S2 t'') S eqS2tS2t'' eqS2t''S
                    eqM1 : equal M1 (T1+S)
                    eqM1 = settaut6 subSM1
                    subM1S2t' : sub M1 (S2 t')
                    subM1S2t' x xM1 = 
                        let alts : el x (S2 t) \/ x==t
                            alts = 
                                let alts' : el x T1 \/ el x S
                                    alts' = ((comprLem x).1 (eqM1.1 x xM1)).1
                                in orE alts' 
                                   (\xT1 -> orI2 (singlT1 x t xT1 tT1))
                                   (\xS -> orI1 (eqS2tS.2 x xS))
                            alt1 : el x (S2 t) -> R2 x t' 
                            alt1 xS2t = tra x t t' (((comprLem x).1 xS2t).1.1) tM2 t'M2 (((comprLem x).1 xS2t).1.2) R2tt'
                            xM2 : el x M2
                            xM2 = subM1M2 x xM1
                            alt2 : x == t -> R2 x t' 
                            alt2 eqxt = eqtrans1 hGSs.2 xM2 tM2 t'M2 eqxt R2tt'
                            lem : R2 x t' -> el x (S2 t')
                            lem R2xt' =
                                let ext : (y:X) -> y==x -> el y M2 /\ R2 y t'
                                    ext y eqyx = 
                                        let yM2 : el y M2
                                            yM2 = M2.2 x y (eqXsym y x eqyx) xM2
                                        in (yM2, eqtrans1 hGSs.2 yM2 xM2 t'M2 eqyx R2xt')
                                in (comprLem x).2 ((xM2, R2xt'), ext)
                        in orE alts (\l -> lem (alt1 l)) (\r -> lem (alt2 r))
                    subS2t'M1 : sub (S2 t') M1
                    subS2t'M1 x xS2t' = 
                        let xM2 : el x M2
                            xM2 = ((comprLem x).1 xS2t').1.1
                            alts : x==t \/ (R2 x t \/ R2 t x)
                            alts = orE (LEM (x==t)) 
                                   (\eqxt -> orI1 eqxt)
                                   (\neqxt -> orI2 (notand.1 (((contrap (hGSs.2.1.1.1 x t xM2 tM2).2.2) neqxt))))
                            alt1 : x==t -> el x M1
                            alt1 eqxt = subT1M1 x (T1.2 t x (eqXsym x t eqxt) tT1)
                            alt2 : R2 x t -> el x M1
                            alt2 R2xt = 
                                let ext : (y:X) -> y==x -> el y M2 /\ R2 y t
                                    ext y eqyx = 
                                        let yM2 : el y M2
                                            yM2 = M2.2 x y (eqXsym y x eqyx) xM2
                                            R2yt : R2 y t
                                            R2yt = eqtrans1 hGSs.2 yM2 xM2 tM2 eqyx R2xt 
                                        in (yM2, R2yt)
                                    xS2t : el x (S2 t)
                                    xS2t = (comprLem x).2 ((xM2, R2xt), ext)
                                in subSM1 x (eqS2tS.1 x xS2t)
                            alt3 : R2 t x -> False
                            alt3 R2tx = 
                                let tri : trich M2 R2
                                    tri = hGSs.2.1.1.1
                                    nxS2t : Not (el x (S2 t))
                                    nxS2t xS2t = ((tri x t xM2 tM2).1.1 ((comprLem x).1 xS2t).1.2).1 R2tx
                                    nxS : Not (el x S)
                                    nxS xS = nxS2t (eqS2tS.2 x xS)
                                    xM2 : el x M2
                                    xM2 = ((comprLem x).1 xS2t').1.1
                                    xT2 : el x T2
                                    xT2 = 
                                        let ext : (y:X) -> y==x -> el y M2 /\ Not (el y S)
                                            ext y eqyx = (M2.2 x y (eqXsym y x eqyx) xM2, \yS -> nxS (S.2 y x eqyx yS))
                                        in (comprLem x).2 ((xM2,nxS), ext)
                                    nxSinglt : Not (el x (Singl t))
                                    nxSinglt xSinglt = 
                                        let eqxt : x==t
                                            eqxt = ((comprLem x).1 xSinglt).1
                                        in ((tri t x tM2 xM2).1.1 R2tx).2 (eqXsym x t eqxt)
                                    xT2' : el x T2'
                                    xT2' = 
                                        let ext : (y:X) -> y==x -> el y T2 /\ Not (el y (Singl t))
                                            ext y eqyx = (T2.2 x y (eqXsym y x eqyx) xT2, \ySinglt -> nxSinglt ((Singl t).2 y x eqyx ySinglt))
                                        in (comprLem x).2 ((xT2, nxSinglt), ext)
                                    R2xt' : R2 x t'
                                    R2xt' = ((comprLem x).1 xS2t').1.2
                                    nR2xt' : Not (R2 x t')
                                    nR2xt' = minimalLem2 M2 R2 (T2',neT2') hGSs.2.1 subT2'M2 x xT2'
                                in nR2xt' R2xt'
                        in orE alts alt1 
                               (\r -> orE r alt2 
                                      (\r' -> whenFalse (el x M1) (alt3 r')))
                    eqS2t'M1 : equal (S2 t') M1
                    eqS2t'M1 = (subS2t'M1, subM1S2t')
                in orI2 (orI2 (t', (t'M2, eqS2t'M1)))
           in orE (LEM (equal T2 (Singl t))) casebi casebii
      in orE (LEM (equal T1 O)) casea caseb
     case2 : Exists (\x -> el x M1 /\ 
              Exists (\x1 -> el x1 (S1 x) /\ equal (S1 x1) M2))
             -> Goal
     case2 h = let x = h.1
                   x1 = h.2.2.1
                   x1S1x = h.2.2.2.1
                   x1iso = h.2.2.2.2
                   x1M1 = ((comprLem x1).1 x1S1x).1.1
               in orI2 (orI1 (x1, (x1M1, x1iso)))
     case3 : Exists (\x -> el x M1 /\ equal (S1 x) M2) -> Goal
     case3 h = orI2 (orI1 h)
 in orE (taut5' (Lemma hGSs)) (case1) (\h -> orE h case2 case3)


iseg : (M1,M2:Pow) -> (R1,R2:Rel X) -> Set
iseg M1 M2 R1 R2 = Exists (\z -> el z M2 /\ equal (IS M2 z R2) M1)

Step5' (M1,M2:Pow) (R1,R2:Rel X) : GS M1 R1 /\ GS M2 R2 ->
    Not (iseg M2 M1 R2 R1) -> equal M1 M2 \/ iseg M1 M2 R1 R2
Step5' {M1} {M2} {R1} {R2} hGSs nisegM2M1 = 
    orE (Step5 hGSs)
        (\eqM1M2 -> orI1 eqM1M2)
        (\h -> orE h 
               (\isegM2M1 -> orI2 (whenFalse (iseg M1 M2 R1 R2) 
                                   (nisegM2M1 isegM2M1))) 
               (\isegM1M2 -> orI2 isegM1M2))

R : Rel X
R a b = 
    Exists (\Ma -> 
    Exists (\Da -> GS Ma (dRel Da) /\ el a Ma /\
            ((Mb:Pow) -> (Db:DRel X) -> 
             GS Mb (dRel Db) -> el b Mb ->
             iseg Ma Mb (dRel Da) (dRel Db)
            )))

(<) = R

GE : X -> Set
GE x = Exists (\U -> el x U /\ GS U R)

LG : Pow
LG = compr (\x -> GE x)

taut7' (P,Q,S:Set) : Not (P /\ Q /\ S) -> P -> Q -> Not S
taut7' {P}{Q}{S} h p q = 
    let lem1 : Not (P /\ Q) \/ Not S
        lem1 = notand''.1 h
    in orE lem1 
           (\npandq -> whenFalse (Not S) (npandq (p,q))) 
           (\ns -> ns)

taut7 (a,b:X) : Not (R a b) -> 
    (Ma:Pow) -> (Da:DRel X) -> GS Ma (dRel Da) -> el a Ma ->
     Exists (\Mb -> Exists (\Db -> GS Mb (dRel Db) /\ el b Mb /\
                                   Not (iseg Ma Mb (dRel Da) (dRel Db))))
taut7 {a}{b} nRab Ma Da gsMa aMa = 
    let lem1 = notex2all nRab
        lem2 = notex2all (lem1 Ma)
        lem3 : Not (GS Ma (dRel Da) /\ el a Ma /\
                    ((Mb:Pow) -> (Db:DRel X) -> 
                     GS Mb (dRel Db) -> el b Mb ->
                     iseg Ma Mb (dRel Da) (dRel Db)))
        lem3 = lem2 Da
        lem4 : Not ((Mb:Pow) -> (Db:DRel X) -> 
                     GS Mb (dRel Db) -> el b Mb ->
                     iseg Ma Mb (dRel Da) (dRel Db))
        lem4 = taut7' lem3 gsMa aMa
        lem5 = notall2ex lem4
        lem6 = notall2ex (lem5.2)
        Mb = lem5.1
        Db = lem6.1
        lem7 : GS Mb (dRel Db) /\ el b Mb /\
             Not (iseg Ma Mb (dRel Da) (dRel Db))
        lem7 = notImp' lem6.2
    in (Mb, (Db, lem7))

rRel : Rel X -> DRel X
rRel RR x = let P : Pred X
                P = RR x
                R : X -> Bool -> Set
                R y b = T b <-> P y
                lem1 : (y:X) -> Exists (\b -> R y b)
                lem1 y = orE (LEM (P y))
                         (\h -> (true, (\h' -> h, \h' -> tt)))
                         (\h -> (false, (\h' -> whenFalse (P y) h', 
                                         \h' -> h h')))
                choicef : Exists (\f -> (y:X) -> R y (f y))
                choicef = IntAC R lem1
            in choicef.1

RiffD : (RR:Rel X) -> (x,y:X) -> dRel (rRel RR) x y <-> RR x y
RiffD RR x = let P : Pred X
                 P = RR x
                 R : X -> Bool -> Set
                 R y b = T b <-> P y
                 lem1 : (y:X) -> Exists (\b -> R y b)
                 lem1 y = orE (LEM (P y))
                          (\h -> (true, (\h' -> h, \h' -> tt)))
                          (\h -> (false, (\h' -> whenFalse (P y) h', 
                                          \h' -> h h')))
                 choicef : Exists (\f -> (y:X) -> R y (f y))
                 choicef = IntAC R lem1
             in choicef.2

ISLem10 (M1,M2:Pow)(R1,R2:Rel X) : iseg M1 M2 R1 R2 -> sub M1 M2
ISLem10 {M1}{M2}{R1}{R2} isegM1M2 x xM1 = 
    let z : X
        z = isegM1M2.1
        subM1ISM2 : sub M1 (IS M2 z R2)
        subM1ISM2 = isegM1M2.2.2.2
    in ((comprLem x).1 (subM1ISM2 x xM1)).1.1

ISLem7 (M1,M2:Pow)(R1,R2:Rel X) : trich M2 R2 ->
                                  iseg M1 M2 R1 R2 -> iseg M2 M1 R2 R1 ->
                                  False
ISLem7 {M1}{M2}{R1}{R2} tri isegM1M2 isegM2M1 = 
    let z : X
        z = isegM1M2.1
        zM2 : el z M2
        zM2 = isegM1M2.2.1
        nzM1 : Not (el z M1)
        nzM1 zM1 = let zISM2 : el z (IS M2 z R2)
                       zISM2 = isegM1M2.2.2.2 z zM1
                       R2zz : R2 z z
                       R2zz = ((comprLem z).1 zISM2).1.2
                   in ((tri z z zM2 zM2).1.1 R2zz).2 (eqXref z)
        subM2M1 : sub M2 M1
        subM2M1 = ISLem10 {M2}{M1}{R2}{R1} isegM2M1
        zM1 : el z M1
        zM1 = subM2M1 z zM2
    in nzM1 zM1

ISLem6 (U:Pow)(R:Rel X) : trich U R -> iseg U U R R -> False
ISLem6 tri is = ISLem7 tri is is

GSlem1 : (U:Pow) -> (RR:Rel X) -> GS U RR <-> GS U (dRel (rRel RR))
GSlem1 U RR =
    let lem : (R1,R2:Rel X) -> ((x,y:X) -> R1 x y <-> R2 x y) -> 
                               GS U R2 -> GS U R1
        lem R1 R2 RiffD' gs = 
            let r2r1 : (a,b:X) -> R2 a b -> R1 a b
                r2r1 a b Rab = (RiffD' a b).2 Rab
                nr2r1 : (a,b:X) -> Not (R2 a b) -> Not (R1 a b)
                nr2r1 a b nRab Dab = nRab ((RiffD' a b).1 Dab)
                r1r2 : (a,b:X) -> R1 a b -> R2 a b
                r1r2 a b Dab = (RiffD' a b).1 Dab
                nr1r2 : (a,b:X) -> Not (R1 a b) -> Not (R2 a b)
                nr1r2 a b nDab Rab = nDab ((RiffD' a b).2 Rab)
                tri : trich U R2
                tri = gs.1.1.1
                tri' : trich U R1
                tri' a b aU bU = 
                    ((\Dab -> (nr2r1 b a ((tri a b aU bU).1.1 (r1r2 a b Dab)).1,
                               ((tri a b aU bU).1.1 (r1r2 a b Dab)).2),
                      \hypo -> r2r1 a b 
                               ((tri a b aU bU).1.2 (nr1r2 b a hypo.1, hypo.2))),
                     (\eqab -> (nr2r1 a b ((tri a b aU bU).2.1 eqab).1, 
                                nr2r1 b a ((tri a b aU bU).2.1 eqab).2), 
                      \hypo -> (tri a b aU bU).2.2 
                               (nr1r2 a b hypo.1, nr1r2 b a hypo.2)))
                tra : trans U R2
                tra = gs.1.1.2
                tra' : trans U R1
                tra' a b c aU bU cU Dab Dbc = 
                    r2r1 a c (tra a b c aU bU cU (r1r2 a b Dab) (r1r2 b c Dbc))
                hl : (V:Pow') -> sub V.1 U -> hasLeast V.1 R2
                hl = gs.1.2
                hl' : (V:Pow') -> sub V.1 U -> hasLeast V.1 R1
                hl' V subVU = 
                    let a = (hl V subVU).1
                        aU = (hl V subVU).2.1
                        lem = (hl V subVU).2.2
                    in (a, (aU, \b bU -> nr2r1 b a (lem b bU)))
                ag : (a:X) -> el a U -> (ne:nonempty (C (IS U a R2))) -> 
                              (==) a (gamma (C (IS U a R2), ne))
                ag = gs.2
                iss : (a:X) -> el a U -> 
                               equal (IS U a R2) (IS U a R1)
                iss a aU = 
                    let sub1 : sub (IS U a R2) (IS U a R1)
                        sub1 x xISR2 = 
                            let xU : el x U
                                xU = ((comprLem x).1 xISR2).1.1
                                Dxa : R1 x a
                                Dxa = r2r1 x a ((comprLem x).1 xISR2).1.2
                                ext : (y:X) -> y==x -> el y U /\ R1 y a
                                ext y eqyx = 
                                    let yU : el y U
                                        yU = U.2 x y (eqXsym y x eqyx) xU
                                    in (yU, eqtrans1' tri' tra' yU xU aU eqyx Dxa)
                            in (comprLem x).2 ((xU, Dxa), ext)
                        sub2 : sub (IS U a R1) (IS U a R2) 
                        sub2 x xISR2 = 
                            let xU : el x U
                                xU = ((comprLem x).1 xISR2).1.1
                                Dxa : R2 x a
                                Dxa = r1r2 x a ((comprLem x).1 xISR2).1.2
                                ext : (y:X) -> y==x -> el y U /\ R2 y a
                                ext y eqyx = 
                                    let yU : el y U
                                        yU = U.2 x y (eqXsym y x eqyx) xU
                                    in (yU, eqtrans1' tri tra yU xU aU eqyx Dxa)
                            in (comprLem x).2 ((xU, Dxa), ext)
                    in (sub1, sub2)
                ag' : (a:X) -> el a U -> 
                               (ne:nonempty (C (IS U a R1))) -> 
                               a == (gamma (C (IS U a R1), ne))
                ag' a aU ne = 
                    let issC : equal (C (IS U a R2)) (C (IS U a R1))
                        issC = CLem1 (iss a aU)
                        ne' : nonempty (C (IS U a R2))
                        ne' = (ne.1, issC.2 ne.1 ne.2)
                        eqag : a == (gamma (C (IS U a R2), ne'))
                        eqag = ag a aU ne'
                        eqgg : (gamma (C (IS U a R2), ne')) == 
                                 (gamma (C (IS U a R1), ne))
                        eqgg = AC2gamma.2.2 (C (IS U a R2), ne') 
                               (C (IS U a R1), ne) issC
                    in eqXtra a (gamma (C (IS U a R2), ne')) 
                           (gamma (C (IS U a R1), ne)) eqag eqgg
            in (((tri', tra'), hl'), ag')
        lem1 : GS U RR -> GS U (dRel (rRel RR))
        lem1 gs = lem (dRel (rRel RR)) RR (RiffD RR) gs
        RiffD' : (x,y:X) -> RR x y <-> dRel (rRel RR) x y
        RiffD' x y = (((RiffD RR) x y).2, ((RiffD RR) x y).1)
        lem2 : GS U (dRel (rRel RR)) -> GS U RR
        lem2 gs = lem RR (dRel (rRel RR)) RiffD' gs
    in (lem1, lem2)

ISLem9 (M1,M2,M3:Pow)(R1,R2,R3:Rel X) : 
       GS M1 R1 -> GS M2 R2 -> GS M3 R3 ->
       iseg M1 M2 R1 R2 -> iseg M2 M3 R2 R3 -> iseg M1 M3 R1 R3
ISLem9 {M1}{M2}{M3}{R1}{R2}{R3} gsM1R1 gsM2R2 gsM3R3 isegM1M2 isegM2M3 = 
    let z2 : X
        z2 = isegM1M2.1
        z2M2 : el z2 M2
        z2M2 = isegM1M2.2.1
        subM2M3 : sub M2 M3
        subM2M3 = ISLem10 {M2}{M3}{R2}{R3} isegM2M3
        z2M3 : el z2 M3
        z2M3 = subM2M3 z2 z2M2
        eqISM2M1 : equal (IS M2 z2 R2) M1
        eqISM2M1 = isegM1M2.2.2
        nz2M1 : Not (el z2 M1)
        nz2M1 = (contrap (eqISM2M1.2 z2)) (ISLem4 gsM2R2 z2 z2M2)
        Alts : equal M1 M3 \/ (iseg M3 M1 R3 R1 \/ iseg M1 M3 R1 R3)
        Alts = Step5 (gsM1R1, gsM3R3)
        lem1 : sub M3 M1 -> False
        lem1 h = nz2M1 (h z2 z2M3)
        alt1 : equal M1 M3 -> False
        alt1 eqM1M3 = lem1 eqM1M3.2
        alt2 : iseg M3 M1 R3 R1 -> False
        alt2 h = lem1 (ISLem10 {M3}{M1}{R3}{R1} h)
    in orE Alts
           (\l -> whenFalse (iseg M1 M3 R1 R3) (alt1 l))
           (\r -> orE r 
                  (\rl -> whenFalse (iseg M1 M3 R1 R3) (alt2 rl))
                  (\rl -> rl))

ISLem11 (M1,M2:Pow)(R:Rel X) : trich M2 R -> trans M2 R -> sub M1 M2 -> (x:X) -> el x M2 -> sub (IS M1 x R) (IS M2 x R)
ISLem11 {M1}{M2}{(<')} triM2 traM2 subM1M2 x xM2 a aISM1 = 
    let aM1 : el a M1
        aM1 = ((comprLem a).1 aISM1).1.1
        aM2 : el a M2
        aM2 = subM1M2 a aM1
        Rax : a <' x
        Rax = ((comprLem a).1 aISM1).1.2
        ext : (b:X) -> b==a -> el b M2 /\ b <' x
        ext b eqba = let bM2 : el b M2
                         bM2 = M2.2 a b (eqXsym b a eqba) aM2
                     in (bM2, eqtrans1' triM2 traM2 bM2 aM2 xM2 eqba Rax)
    in (comprLem a).2 ((aM2, Rax), ext)

trich1R : (a,b:X) -> (a<b) -> Not (b<a) /\ Not (a==b)
trich1R a b Rab =             
    let lem2 : Not (a==b)
        lem2 eqab = 
            let Ma : Pow
                Ma = Rab.1
                Da : DRel X
                Da = Rab.2.1
                gsMaDa : GS Ma (dRel Da)
                gsMaDa = Rab.2.2.1.1
                aMa : el a Ma
                aMa = Rab.2.2.1.2
                bMa : el b Ma
                bMa = Ma.2 a b eqab aMa
                isegMaMa : iseg Ma Ma (dRel Da) (dRel Da)
                isegMaMa = Rab.2.2.2 Ma Da gsMaDa bMa
            in ISLem6 {Ma} {dRel Da} gsMaDa.1.1.1 isegMaMa
        lem1 : Not (b<a)
        lem1 Rba = 
            let Lb : Pow
                Lb = Rba.1
                Db : DRel X
                Db = Rba.2.1
                gsLbDb : GS Lb (dRel Db)
                gsLbDb = Rba.2.2.1.1
                bLb : el b Lb
                bLb = Rba.2.2.1.2
                Ma : Pow
                Ma = Rab.1
                Da : DRel X
                Da = Rab.2.1
                gsMaDa : GS Ma (dRel Da)
                gsMaDa = Rab.2.2.1.1
                aMa : el a Ma
                aMa = Rab.2.2.1.2
                isegLbMa : iseg Lb Ma (dRel Db) (dRel Da)
                isegLbMa = Rba.2.2.2 Ma Da gsMaDa aMa
                isegMaLb : iseg Ma Lb (dRel Da) (dRel Db)
                isegMaLb = Rab.2.2.2 Lb Db gsLbDb bLb
            in ISLem7 gsMaDa.1.1.1 isegLbMa isegMaLb
    in (lem1, lem2)

Step7I' : (a,b:X) -> el a LG -> el b LG -> (a<b) <-> Not (b<a) /\ Not (a==b)
Step7I' a b aLG bLG =
    let aGE : GE a
        aGE = ((comprLem a).1 aLG).1
        bGE : GE b
        bGE = ((comprLem b).1 bLG).1
        Ma : Pow
        Ma = aGE.1
        Mb : Pow
        Mb = bGE.1
        aMa : el a Ma
        aMa = aGE.2.1
        bMb : el b Mb
        bMb = bGE.2.1
        gsMaR : GS Ma R
        gsMaR = aGE.2.2
        gsMbR : GS Mb R
        gsMbR = bGE.2.2
        triMb : trich Mb R
        triMb = gsMbR.1.1.1
        gsMbR' : GS Mb (dRel (rRel R))
        gsMbR' = (GSlem1 Mb R).1 gsMbR
        lem1 : (a<b) -> Not (b<a) /\ Not (a==b)
        lem1 = trich1R a b
        lem2 : Not (b<a) /\ Not (a==b) -> (a<b)
        lem2 hypo =
            let taut7app = taut7 hypo.1 Mb (rRel R) gsMbR' bMb
		La : Pow
                La = taut7app.1
                Da : DRel X
                Da = taut7app.2.1
                gsLaDa : GS La (dRel Da)
                gsLaDa = taut7app.2.2.1.1
                aLa : el a La
                aLa = taut7app.2.2.1.2
                lem20 : Not (iseg Mb La (dRel (rRel R)) (dRel Da))
                lem20 = taut7app.2.2.2
                lem21 : equal La Mb \/
                        iseg La Mb (dRel Da) (dRel (rRel R))
                lem21 = Step5' (gsLaDa,gsMbR') lem20
                lem22 : sub La Mb
                lem22 = orE lem21 
                        (\l -> l.1)
                        (\r -> ISLem10 {La}{Mb}{(dRel Da)}{(dRel (rRel R))} r)
                aMb : el a Mb
                aMb = lem22 a aLa
            in (triMb a b aMb bMb).1.2 hypo
    in (lem1, lem2)

Step7I : trich LG (<)
Step7I a b aLG bLG =
    let lem' : Not (a<b) <-> (b<a) \/ (a==b)
        lem' = (\h -> notand.1 ((contrap (Step7I' a b aLG bLG).2) h), 
                \h -> (contrap (Step7I' a b aLG bLG).1) (notand.2 h))
        lem3 : a==b -> Not (a<b) /\ Not (b<a)
        lem3 eqab =
            let nRab : Not (a<b)
                nRab = lem'.2 (orI2 eqab)
                nRba : Not (b<a)
                nRba Rba = ((Step7I' b a bLG aLG).1 Rba).2 (eqXsym a b eqab)
            in (nRab, nRba)
        lem4 : Not (a<b) /\ Not (b<a) -> a==b
        lem4 h = orE (lem'.1 h.1)
                  (\Rba -> whenFalse (a==b) (h.2 Rba))
                  (\eqab -> eqab)
    in (Step7I' a b aLG bLG, (lem3, lem4))

transR : (U:Pow) -> trans U R
transR U a b c aU bU cU Rab Rbc = 
    let Ma = Rab.1
        Da = Rab.2.1
        gsMaDa = Rab.2.2.1.1
        aMa = Rab.2.2.1.2
        Mb = Rbc.1
        Db = Rbc.2.1
        gsMbDb = Rbc.2.2.1.1
        bMb = Rbc.2.2.1.2
        lem1 : (Mc:Pow) -> (Dc:DRel X) -> GS Mc (dRel Dc) -> el c Mc ->
                           iseg Ma Mc (dRel Da) (dRel Dc)
        lem1 Mc Dc gsMcDc cMc = ISLem9 {Ma}{Mb}{Mc}{dRel Da}{dRel Db}{dRel Dc}
                                gsMaDa gsMbDb gsMcDc
                                (Rab.2.2.2 Mb Db gsMbDb bMb)
                                (Rbc.2.2.2 Mc Dc gsMcDc cMc)
    in (Ma, (Da, ((gsMaDa, aMa), lem1)))

Step7II' : trans LG R
Step7II' = transR LG

Step7II : linear LG R
Step7II = (Step7I, Step7II')

Step7III : wellOrdered LG R
Step7III = 
 (Step7II, 
  \L1 subL1LG ->
   let a : X
       a = L1.2.1
       aL1 : el a L1.1
       aL1 = L1.2.2
       L2 : Pow
       L2 = compr (\x -> el x (IS L1.1 a R) \/ x==a)
       aLG : el a LG
       aLG = subL1LG a aL1
       aGE : GE a
       aGE = ((comprLem a).1 aLG).1
       M1 : Pow
       M1 = aGE.1
       aM1 : el a M1
       aM1 = aGE.2.1
       gsM1R : GS M1 R
       gsM1R = aGE.2.2
       gsM1R' : GS M1 (dRel (rRel R))
       gsM1R' = (GSlem1 M1 R).1 gsM1R
       subL2M1 : sub L2 M1
       subL2M1 x xL2 = 
           let lem1 : el x (IS L1.1 a R) -> el x M1
               lem1 h = 
                   let Rxa : x<a
                       Rxa = ((comprLem x).1 h).1.2
                       Mx : Pow
                       Mx = Rxa.1
                       xMx : el x Mx
                       xMx = Rxa.2.2.1.2
                       subMxM1 : sub Mx M1
                       subMxM1 = ISLem10 {Mx}{M1}{R}{dRel (rRel R)} (Rxa.2.2.2 M1 (rRel R) gsM1R' aM1)
                   in subMxM1 x xMx
               lem2 : x==a -> el x M1
               lem2 h = M1.2 a x (eqXsym x a h) aM1
           in orE ((comprLem x).1 xL2).1 lem1 lem2
       subL2L1 : sub L2 L1.1
       subL2L1 x xL2 =
           let lem1 : el x (IS L1.1 a R) -> el x L1.1
               lem1 h = ((comprLem x).1 h).1.1
               lem2 : x==a -> el x L1.1
               lem2 h = L1.1.2 a x (eqXsym x a h) aL1
           in orE ((comprLem x).1 xL2).1 lem1 lem2
       ext : (y:X) -> y==a -> el y (IS L1.1 a R) \/ y==a
       ext y eqya = orI2 eqya
       aL2 : el a L2
       aL2 = (comprLem a).2 (orI2 (eqXref a), ext)
       neL2 : nonempty L2
       neL2 = (a, aL2)
       hlL2R : hasLeast L2 R
       hlL2R = gsM1R.1.2 (L2,neL2) subL2M1
       min : X
       min = hlL2R.1
       minL2 : el min L2
       minL2 = hlL2R.2.1
       minLG : el min LG
       minLG = subL1LG min (subL2L1 min minL2)
       hlL1R : hasLeast L1.1 R
       hlL1R = let lem1 : el min L1.1
                   lem1 = subL2L1 min minL2
                   lem2 : (b:X) -> el b L1.1 -> Not (b<min)
                   lem2 b bL1 Rbmin = 
                       let alt : min<a \/ a==min
                           alt = notand.1 ((contrap (Step7I a min aLG minLG).1.2) (hlL2R.2.2 a aL2))
                           bLG : el b LG
                           bLG = subL1LG b bL1
                           Rba : b<a
                           Rba = orE alt
                                 (\Rmina -> Step7II' b min a bLG minLG aLG 
                                            Rbmin Rmina)
                                 (\eqamin -> eqtrans2' Step7I Step7II' aLG 
                                             minLG bLG eqamin Rbmin)
                           ext : (y:X) -> y==b -> el y L1.1 /\ y<a
                           ext y eqyb = let yL1 : el y L1.1
                                            yL1 = L1.1.2 b y (eqXsym y b eqyb) bL1
                                            yLG : el y LG
                                            yLG = subL1LG y yL1
                                        in (yL1, eqtrans1' Step7I Step7II'
                                                   yLG bLG aLG eqyb Rba)
                           bIS : el b (IS L1.1 a R)
                           bIS = (comprLem b).2 ((bL1, Rba), ext)
                           ext' : (y:X) -> y==b -> el y (IS L1.1 a R) \/ y==a
                           ext' y eqyb = orI1 ((IS L1.1 a R).2 b y (eqXsym y b eqyb) bIS)
                           bL2 : el b L2
                           bL2 = (comprLem b).2 (orI1 bIS, ext')
                           nRbmin : Not (b<min)
                           nRbmin = hlL2R.2.2 b bL2
                       in nRbmin Rbmin
               in (min, (lem1, lem2))
   in hlL1R)

Step7IV : GS LG R
Step7IV = 
 let lem1 : (a:X) -> el a LG -> (ne:nonempty (C (IS LG a R))) -> 
                     a == (gamma (C (IS LG a R), ne))
     lem1 a aLG neCA = 
      let aGE = ((comprLem a).1 aLG).1
          Ma = aGE.1
          aMa : el a Ma
          aMa = aGE.2.1
          MaGS : GS Ma R
          MaGS = aGE.2.2
          B = IS Ma a R
          A = IS LG a R
          eqAB : equal A B
          eqAB = let subAB : sub A B
                     subAB x xA = 
                         let Rxa : x<a
                             Rxa = ((comprLem x).1 xA).1.2
                             Mx : Pow
                             Mx = Rxa.1
                             xMx : el x Mx
                             xMx = Rxa.2.2.1.2
                             gsMa : GS Ma (dRel (rRel R))
                             gsMa = (GSlem1 Ma R).1 MaGS
                             subMxMa : sub Mx Ma
                             subMxMa = ISLem10 {Mx}{Ma}{R}{dRel (rRel R)} (Rxa.2.2.2 Ma (rRel R) gsMa aMa)
                             xMa : el x Ma
                             xMa = subMxMa x xMx
                             ext : (y:X) -> y==x -> el y Ma /\ y<a
                             ext y eqyx =
                                 let yMa : el y Ma
                                     yMa = Ma.2 x y (eqXsym y x eqyx) xMa
                                 in (yMa, eqtrans1 MaGS yMa xMa aMa eqyx Rxa)
                         in (comprLem x).2 ((xMa,Rxa),ext)
                     subBA : sub B A
                     subBA x xB = 
                         let xMa : el x Ma
                             xMa = ((comprLem x).1 xB).1.1
                             xGE : GE x
                             xGE = (Ma, (xMa, MaGS))
                             ext : (y:X) -> y==x -> GE y
                             ext y eqyx = (xGE.1, (xGE.1.2 x y (eqXsym y x eqyx) xGE.2.1, xGE.2.2))
                             xLG : el x LG
                             xLG = (comprLem x).2 (xGE,ext)
                             Rxa : x<a
                             Rxa = ((comprLem x).1 xB).1.2
                             ext' : (y:X) -> y==x -> el y LG /\ y<a
                             ext' y eqyx = 
                                 let yLG : el y LG
                                     yLG = LG.2 x y (eqXsym y x eqyx) xLG
                                     yMa : el y Ma
                                     yMa = Ma.2 x y (eqXsym y x eqyx) xMa
                                 in (yLG, eqtrans1 MaGS yMa xMa aMa eqyx Rxa)
                         in (comprLem x).2 ((xLG, Rxa), ext')
                 in (subAB, subBA)
          eqCACB : equal (C A) (C B)
          eqCACB = CLem1 eqAB
          neCB : nonempty (C B)
          neCB = let w=neCA.1 in (w, eqCACB.1 w neCA.2)
          CA' : Pow'
          CA' = (C A, neCA)
          CB' : Pow'
          CB' = (C B, neCB)
          eq'CACB : equal' CA' CB'
          eq'CACB = eqCACB
          eqagCB' : a == gamma CB'
          eqagCB' = MaGS.2 a aMa neCB
          eqgCAgCB : gamma CA' == gamma CB'
          eqgCAgCB = AC2gamma.2.2 CA' CB' eq'CACB
          eqagCA : a == gamma CA'
          eqagCA = eqXtra a (gamma CB') (gamma CA') eqagCB' 
                   (eqXsym (gamma CA') (gamma CB') eqgCAgCB)
      in eqagCA
 in (Step7III, lem1)

Step7V : (x:X) -> el x LG
Step7V x = 
 orE (LEM (el x LG))
 (\h -> h)
 (\nxLG -> 
  let xCLG : el x (C LG)
      xCLG = let ext : (y:X) -> y==x -> Not (el y LG)
                 ext y eqyx yLG = nxLG (LG.2 y x eqyx yLG)
             in (comprLem x).2 (nxLG, ext)
      neCLG : nonempty (C LG)
      neCLG = (x, xCLG)
      m : X
      m = gamma (C LG, neCLG)
      L' : Pow
      L' = compr (\x -> el x LG \/ x==m)
      mL' : el m L'
      mL' = let ext : (y:X) -> y==m -> el y LG \/ y==m
                ext y eqym = orI2 eqym
            in (comprLem m).2 (orI2 (eqXref m), ext)
      neqm : (x:X) -> el x L' -> Not (x==m) -> el x LG
      neqm y yL' neqym = orE ((comprLem y).1 yL').1
                         (\yLG -> yLG)
                         (\eqym -> whenFalse (el y LG) (neqym eqym))
      mCLG : el m (C LG)
      mCLG = AC2gamma.2.1 (C LG, neCLG)
      nmLG : Not (el m LG)
      nmLG = ((comprLem m).1 mCLG).1
      lemII : (a,b:X) -> el a L' -> el b L' -> Not (a==m) -> b==m -> a<b
      lemII a b aL' bL' neqam eqbm = 
          let aLG : el a LG
              aLG = neqm a aL' neqam
              aGE : GE a
              aGE = ((comprLem a).1 aLG).1
              Ma : Pow
              Ma = aGE.1
              gsMa : GS Ma (dRel (rRel R))
              gsMa = (GSlem1 Ma R).1 aGE.2.2
              aMa : el a Ma
              aMa = aGE.2.1
              lem : (Mm:Pow) -> (Dm:DRel X) -> GS Mm (dRel Dm) -> el b Mm ->
                                iseg Ma Mm (dRel (rRel R)) (dRel Dm)
              lem Mm Dm gsMm bMm = 
                  let Goal = iseg Ma Mm (dRel (rRel R)) (dRel Dm)
                      mMm : el m Mm
                      mMm = Mm.2 b m eqbm bMm
                  in orE (LEM Goal)
                     (\goal -> goal)
                     (\ngoal -> 
                          let lem1 : equal Mm Ma \/ iseg Mm Ma (dRel Dm) (dRel (rRel R))
                              lem1 = Step5' (gsMm, gsMa) ngoal
                              lem2 : sub Mm Ma
                              lem2 = orE lem1 
                                     (\l -> l.1)
                                     (\r -> ISLem10 {Mm}{Ma}{(dRel Dm)}{(dRel (rRel R))} r)
                              mMa : el m Ma
                              mMa = lem2 m mMm
                              mGE : GE m
                              mGE = (Ma, (mMa, aGE.2.2))
                              ext : (y:X) -> y==m -> GE y
                              ext y eqym = ((Ma, (Ma.2 m y (eqXsym y m eqym) mMa, aGE.2.2)))
                              mLG : el m LG
                              mLG = (comprLem m).2 (mGE, ext)
                          in whenFalse Goal (nmLG mLG))
          in (Ma, (rRel R, ((gsMa, aMa), lem)))
      tra : trans L' R
      tra = transR L'
      tri1 : (a,b:X) -> el a L' -> el b L' -> R a b -> Not(R b a) /\ Not (a==b)
      tri1 a b aL' bL' = trich1R a b
      tri2 : (a,b:X) -> el a L' -> el b L' -> Not(R b a) /\ Not (a==b) -> R a b
      tri2 a b aL' bL' hypo = 
          let alt1 : a==m -> b==m -> a<b
              alt1 eqam eqbm = whenFalse (a<b)
                               (hypo.2 
                                (eqXtra a m b eqam (eqXsym b m eqbm)))
              alt2 : Not (a==m) -> b==m -> a<b
              alt2 neqam eqbm = lemII a b aL' bL' neqam eqbm
              alt3 : a==m -> Not (b==m) -> a<b
              alt3 eqam neqbm =
                  let Rba : b<a
                      Rba = lemII b a bL' aL' neqbm eqam
                  in whenFalse (a<b) (hypo.1 Rba)
              alt4 : Not (a==m) -> Not (b==m) -> a<b
              alt4 neqam neqbm = 
                  let aLG : el a LG
                      aLG = neqm a aL' neqam
                      bLG : el b LG
                      bLG = neqm b bL' neqbm
                  in (Step7I' a b aLG bLG).2 hypo
          in orE (LEM (a==m))
             (\eqam -> orE (LEM (b==m))
                       (\eqbm -> alt1 eqam eqbm) 
                       (\neqbm -> alt3 eqam neqbm))
             (\neqam -> orE (LEM (b==m)) 
                       (\eqbm -> alt2 neqam eqbm) 
                       (\neqbm -> alt4 neqam neqbm))
      tri : trich L' R
      tri a b aL' bL' =
          let lem' : Not (a<b) <-> (b<a) \/ (a==b)
              lem' = (\h -> notand.1 ((contrap (tri2 a b aL' bL')) h), 
                      \h -> (contrap (tri1 a b aL' bL')) (notand.2 h))
              lem3 : a==b -> Not (R a b) /\ Not (R b a)
              lem3 eqab =
                  let nRab : Not (a<b)
                      nRab = lem'.2 (orI2 eqab)
                      nRba : Not (b<a)
                      nRba Rba = ((tri1 b a bL' aL') Rba).2 (eqXsym a b eqab)
                  in (nRab, nRba)
              lem4 : Not (R a b) /\ Not (R b a) -> a==b
              lem4 h = orE (lem'.1 h.1)
                       (\Rba -> whenFalse (a==b) (h.2 Rba))
                       (\eqab -> eqab)
          in ((tri1 a b aL' bL',tri2 a b aL' bL'),(lem3,lem4))
      lem1 : (V:Pow') -> sub V.1 L' -> hasLeast V.1 R
      lem1 V subVL' = 
          let alt1 : el m V.1 -> hasLeast V.1 R
              alt1 mV = 
                  let W : Pow
                      W = IS V.1 m R
                      alt11 : nonempty W -> hasLeast V.1 R
                      alt11 neW =
                          let mL' : el m L'
                              mL' = subVL' m mV
                              sub1 : sub W (IS L' m R)
                              sub1 = ISLem11 {V.1}{L'}{R} tri tra subVL' m mL'
                              sub2 : sub (IS L' m R) LG
                              sub2 y yIS =
                                  let yL' : el y L'
                                      yL' = ((comprLem y).1 yIS).1.1
                                      Rym : y<m
                                      Rym = ((comprLem y).1 yIS).1.2
                                      neqym : Not (y==m)
                                      neqym = ((tri y m yL' mL').1.1 Rym).2
                                  in neqm y yL' neqym
                              subWLG : sub W LG
                              subWLG y yW = sub2 y (sub1 y yW)
                              min : X
                              min = minimal LG R (W,neW) Step7III subWLG
                              minW : el min W
                              minW = minimalLem LG R (W,neW) Step7III subWLG
                              minV : el min V.1
                              minV = ((comprLem min).1 minW).1.1
                              lem11 : (b:X) -> el b V.1 -> Not (b<min)
                              lem11 b bV Rbmin = 
                                  let Rminm : min<m
                                      Rminm = ((comprLem min).1 minW).1.2
                                      bL' = subVL' b bV
                                      minL' = subVL' min minV
                                      Rbm : b<m
                                      Rbm = tra b min m bL' minL' mL' Rbmin Rminm
                                      ext : (y:X) -> y==b -> el y V.1 /\ y<m
                                      ext y eqyb = 
                                          let yV : el y V.1
                                              yV = V.1.2 b y (eqXsym y b eqyb) bV
                                              yL' : el y L'
                                              yL' = subVL' y yV
                                          in (yV, eqtrans1' tri tra yL' bL' mL' eqyb Rbm)
                                      bW : el b W
                                      bW = (comprLem b).2 ((bV, Rbm), ext)
                                      nRbmin : Not (b<min)
                                      nRbmin = minimalLem2 LG R (W,neW) Step7III subWLG b bW
                                  in nRbmin Rbmin
                          in (min, (minV, lem11))
                      alt12 : Not (nonempty W) -> hasLeast V.1 R
                      alt12 nneW =
                          let nneW' : (x:X) -> Not (el x W)
                              nneW' = notex2all nneW
                              lem : (b:X) -> el b V.1 -> Not (b<m)
                              lem b bV Rbm = 
                                  let ext : (y:X) -> y==b -> el y V.1 /\ y<m
                                      ext y eqyb = 
                                          let yV : el y V.1
                                              yV  = V.1.2 b y (eqXsym y b eqyb) bV
                                              yL' = subVL' y yV
                                              bL' = subVL' b bV
                                          in (yV, eqtrans1' tri tra yL' bL' mL' eqyb Rbm)
                                      bW : el b W
                                      bW = (comprLem b).2 ((bV,Rbm),ext)
                                  in (nneW' b) bW
                          in (m, (mV, lem))
                  in orE (LEM (nonempty W)) alt11 alt12
              alt2 : Not (el m V.1) -> hasLeast V.1 R
              alt2 nmV = let subVLG : sub V.1 LG
                             subVLG y yV = 
                                 let neqym : Not (y==m)
                                     neqym eqym = nmV (V.1.2 y m eqym yV)
                                 in neqm y (subVL' y yV) neqym
                         in Step7III.2 V subVLG
          in orE (LEM (el m V.1)) alt1 alt2
      lem2 : wellOrdered L' R
      lem2 = ((tri, tra), lem1)
      lem3 : (a:X) -> el a L' -> (ne:nonempty (C (IS L' a R))) -> 
                      a == gamma (C (IS L' a R), ne)
      lem3 a aL' ne = 
          let alt1 : a==m -> sub (IS L' a R) LG
              alt1 eqam y yIS = 
                  let yL' : el y L'
                      yL' = ((comprLem y).1 yIS).1.1
                      Rya : y<a
                      Rya = ((comprLem y).1 yIS).1.2
                      neqya : Not (y==a)
                      neqya = ((tri y a yL' aL').1.1 Rya).2
                      neqym : Not (y==m)
                      neqym eqym = neqya (eqXtra y m a eqym (eqXsym a m eqam))
                  in neqm y yL' neqym
              alt1' : a==m -> sub LG (IS L' a R)
              alt1' eqam y yLG = 
                  let yL' : el y L'
                      yL' = let ext : (z:X) -> z==y -> el z LG \/ z==m
                                ext z eqzy = orI1 (LG.2 y z (eqXsym z y eqzy) yLG)
                            in (comprLem y).2 (orI1 yLG, ext)
                      neqym : Not (y==m)
                      neqym eqym = let mLG : el m LG
                                       mLG = LG.2 y m eqym yLG
                                   in nmLG mLG
                      Rya : y<a
                      Rya = lemII y a yL' aL' neqym eqam
                      ext : (z:X) -> z==y -> el z L' /\ z<a
                      ext z eqzy = 
                          let zL' : el z L'
                              zL' = L'.2 y z (eqXsym z y eqzy) yL'
                          in (zL', eqtrans1' tri tra zL' yL' aL' eqzy Rya)
                  in (comprLem y).2 ((yL', Rya), ext)
              lem1 : a==m -> a == gamma (C (IS L' a R), ne)
              lem1 eqam = 
                  let lem11 : equal LG (IS L' a R)
                      lem11 = (alt1' eqam, alt1 eqam)
                      lem12 : equal (C LG) (C (IS L' a R))
                      lem12 = CLem1 lem11
                      CLG : Pow'
                      CLG = (C LG, neCLG)
                      CIS : Pow'
                      CIS = (C (IS L' a R), ne)
                      lem13 : gamma CLG == m 
                      lem13 = eqXref m
                      lem14 : gamma CLG == gamma CIS
                      lem14 = AC2gamma.2.2 CLG CIS lem12
                      lem15 : a == gamma CLG
                      lem15 = eqXtra a m (gamma CLG) eqam lem13
                  in eqXtra a (gamma CLG) (gamma CIS) lem15 lem14
              alt2 : Not (a==m) -> sub (IS L' a R) LG
              alt2 neqam y yIS = 
                  let yL' : el y L'
                      yL' = ((comprLem y).1 yIS).1.1
                      Ram : a<m
                      Ram = lemII a m aL' mL' neqam (eqXref m)
                      Rya : y<a
                      Rya = ((comprLem y).1 yIS).1.2
                      Rym : y<m
                      Rym = tra y a m yL' aL' mL' Rya Ram
                      neqym : Not (y==m)
                      neqym = ((tri y m yL' mL').1.1 Rym).2
                  in neqm y yL' neqym
              alt2' : Not (a==m) -> equal (IS L' a R) (IS LG a R) 
              alt2' neqam =
                  let sub1 : sub (IS L' a R) (IS LG a R) 
                      sub1 z zISL' = 
                          let Rza : z<a
                              Rza = ((comprLem z).1 zISL').1.2
                              zLG : el z LG
                              zLG = (alt2 neqam) z zISL'
                              ext : (y:X) -> y==z -> el y LG /\ y<a
                              ext y eqyz = 
                                  let yLG = LG.2 z y (eqXsym y z eqyz) zLG
                                      aLG = neqm a aL' neqam
                                  in (yLG, eqtrans1' Step7I Step7II' yLG zLG aLG eqyz Rza)
                          in (comprLem z).2 ((zLG, Rza), ext)
                      sub2 : sub (IS LG a R) (IS L' a R) 
                      sub2 z zISLG =
                          let zLG : el z LG
                              zLG = ((comprLem z).1 zISLG).1.1
                              subLGL' : sub LG L'
                              subLGL' b bLG =
                                  let ext : (y:X) -> y==b -> el y LG \/ y==m
                                      ext y eqyb = orI1 (LG.2 b y (eqXsym y b eqyb) bLG)
                                      bL' : el b L'
                                      bL' = (comprLem b).2 (orI1 bLG, ext)
                                  in bL'
                              zL' : el z L'
                              zL' = subLGL' z zLG
                              Rza : z<a
                              Rza = ((comprLem z).1 zISLG).1.2
                              ext : (y:X) -> y==z -> el y L' /\ y<a
                              ext y eqyz = 
                                  let yLG = LG.2 z y (eqXsym y z eqyz) zLG
                                      yL' = subLGL' y yLG
                                      Rya : y<a
                                      Rya = eqtrans1' tri tra yL' zL' aL' eqyz Rza
                                  in (yL', Rya)
                          in (comprLem z).2 ((zL', Rza), ext)
                  in (sub1, sub2)
              lem2 : Not (a==m) -> a == gamma (C (IS L' a R), ne)
              lem2 neqam =
                  let aLG : el a LG
                      aLG = neqm a aL' neqam
                      eqISs : equal (IS L' a R) (IS LG a R)
                      eqISs = alt2' neqam
                      lem22 : equal (C (IS LG a R)) (C (IS L' a R))
                      lem22 = CLem1 ((alt2' neqam).2, (alt2' neqam).1)
                      ne' : nonempty (C (IS LG a R))
                      ne' = (ne.1, lem22.2 ne.1 ne.2)
                      CISLG : Pow'
                      CISLG = (C (IS LG a R), ne')
                      CISL' : Pow'
                      CISL' = (C (IS L' a R), ne)
                      lem21 : a == gamma CISLG
                      lem21 = Step7IV.2 a aLG ne'
                      lem23 : gamma CISLG == gamma CISL'
                      lem23 = AC2gamma.2.2 CISLG CISL' lem22
                  in eqXtra a (gamma CISLG) (gamma CISL') lem21 lem23
          in orE (LEM (a==m)) lem1 lem2
      lem4 : GS L' R
      lem4 = (lem2,lem3)
      mGE : GE m
      mGE = (L', (mL', lem4))
      ext : (y:X) -> y==m -> GE y
      ext y eqym = (L', (L'.2 m y (eqXsym y m eqym) mL', lem4))
      mLG : el m LG
      mLG = (comprLem m).2 (mGE, ext)
  in whenFalse (el x LG) (nmLG mLG))

