-- The state monad transformer
-- Author :: Ulf Norell
-- Created:: 2004-12-16

-- # import Monad
-- # import Monad.Trans
-- # import Prelude.Proofs
-- # import EqualityReasoning

-- The state monad transformer --------------------------------------------

ReaderT : Set -> (Set -> Set) -> Set -> Set
ReaderT E M A = E -> M A

monadTransReaderT : (E : Set) -> MonadTransformer (ReaderT E)
monadTransReaderT E = struct
    ops    = struct { monad = monad; lift = lift}
    proofs = proofs
  where
    lift : (M : Set -> Set) -> MonadOps M -> (A : Set) -> M A -> ReaderT E M A
    lift M monadM A m = \_ -> m

    return : (M : Set -> Set) -> MonadOps M -> (A : Set) -> A -> ReaderT E M A
    return M monadM A x = \_ -> monadM.return A x

    bind : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
            ReaderT E M A -> (A -> ReaderT E M B) -> ReaderT E M B
    bind M monadM A B m k = \e -> m e >>= \x -> k x e
      where
        (>>=) = monadM.bind A B

    monad : (M : Set -> Set) -> MonadOps M -> MonadOps (ReaderT E M)
    monad M monadM = struct { return = return M monadM; bind = bind M monadM}

    proofs : MonadTransformerLaws (ReaderT E) monad lift
    proofs = struct
          monadLaws  = monadLaws
          liftReturn = \M monadM A x -> fol-plugin( extensionality )
          liftBind   = \M monadM A B m k -> fol-plugin( extensionality )
        where
          monadLaws : (M : Set -> Set) -> (monadM : Monad M) ->
                        MonadLaws (ReaderT E M) (return M monadM.ops)
                                                (bind M monadM.ops)
          monadLaws M monadM = struct
            law1 = \A m         -> fol-plugin( monadM.proofs.law1 A, 
                                               extensionality )
            law2 = \A B x f     -> fol-plugin( monadM.proofs.law2 A B, 
                                               extensionality )
            law3 = \A B C m g h -> fol-plugin( monadM.proofs.law3 A B C, 
                                               extensionality )

-- The Reader monad --------------------------------------------------------

Reader : Set -> Set -> Set
Reader E = ReaderT E Id

monadReader : (E : Set) -> Monad (Reader E)
monadReader E = transformMonad  (ReaderT E) (monadTransReaderT E)
                                Id          monadId

-- Reader monad operations -------------------------------------------------

