-- Monoids
--   There are two approaches::
--     Monoid : Type
--     Monoid = { A : Set; zero : A; ... }
--   and
--     Monoid : (A:Set) -> Type
--     Monoid A = { zero : A; ... }
--   I'm not sure which one I prefer. The second one is closer
--   to Haskell type classes and doesn't require naming of the
--   dictionary and projections::
--     mappend : (m:Monoid) -> List m.A -> m.A
--   so I'll stick with that one for now.
-- Author :: Ulf Norell
-- Created:: 2004-12-14

-- # import Property
-- # import List

IsIdentity : (A : Set) -> A -> (A -> A -> A) -> Prop
IsIdentity A e (+*) = sig
  left   : (x:A) -> (e +* x) == x
  right  : (x:A) -> (x +* e) == x

MonoidLaws : (A:Set) -> A -> (A -> A -> A) -> Prop
MonoidLaws A z (+*) = sig
  id     : IsIdentity A z (+*)
  assoc  : (a,b,c:A) -> ((a +* b) +* c) == (a +* (b +* c))
  

MonoidOps : (A:Set) -> Set
MonoidOps A = sig
  zero   : A
  plus   : A -> A -> A
  

Monoid : (A:Set) -> Type
Monoid A = sig
  ops    : MonoidOps A
  proofs : MonoidLaws A ops.zero ops.plus
  

mappend : (A:Set) -> MonoidOps A -> List A -> A
mappend A mon []      = mon.zero
mappend A mon (x::xs) = mon.plus x (mappend A mon xs)

CommutativeMonoidLaws : (A:Set) -> A -> (A -> A -> A) -> Prop
CommutativeMonoidLaws A z (+*) = sig
  monoidLaws : MonoidLaws A z (+*)
  commute    : (a,b : A) -> (a +* b) == (b +* a)
  

CommutativeMonoid : (A:Set) -> Type
CommutativeMonoid A = sig
  ops    : MonoidOps A
  proofs : CommutativeMonoidLaws A ops.zero ops.plus
  

