{-# OPTIONS -cpp #-}

module IO.Temp where

import System.Environment
import System.Directory

import IO.Temp.MkTemp (mktemp)
import Utils.Monads

slash :: String
#ifdef mingw32_HOST_OS
slash = "\\"
#else
slash = "/"
#endif

getTempDir :: IO FilePath
getTempDir =
    do	home <- (:[]) <$> getEnv "HOME" `catch` \_ -> return []
	getEnv "TEMP" `catch` \_ -> tryOtherDirs (otherDirs home)
    where
	isWritable d =
	    ifM (doesDirectoryExist d)
		(writable <$> getPermissions d)
		(return False)

	tryOtherDirs []	    = return "."
	tryOtherDirs (d:ds) =
	    ifM (isWritable d)
		(return d)
		(tryOtherDirs ds)

#ifdef mingw32_HOST_OS
	otherDirs home = ["C:\\TEMP", "C:\\TMP"] ++ map (++"\\Temporary Files") home
#else
	otherDirs home = ["/tmp","/var/tmp","/usr/local/tmp"] ++ map (++"/tmp") home
#endif

mkTempFile :: String -> IO FilePath
mkTempFile path =
    do	tmp <- getTempDir
	let template = tmp ++ slash ++ path ++ ".XXXXXXXXX"
	p   <- mktemp template
	case p of
	    Nothing -> error $ "mkTempFile: Couldn't create temp file '" ++ template ++ "'"
	    Just p  -> return p

