
module TypeChecker.Monad.Signatures where

import Control.Monad.State
import qualified Data.Set as Set

import Syntax.Internal
import Syntax.Values
import TypeChecker.Monad
import TypeChecker.Errors
import TypeChecker.Monad.Errors
import Utils.Monads
import Utils.Map as Map

getSignatures :: TC Signatures
getSignatures = tcSignatures <$> get

lookupSignature :: Name -> TC [(Name,FieldSource,DefValue)]
lookupSignature x =
    do	mcs <- Map.lookup x <$> getSignatures
	case mcs of
	    Just cs -> return $ sigFields cs
	    Nothing -> failure $ becauseNotASignature (vDef x 0)

lookupSignature_ :: Name -> TC [(Name, DefValue)]
lookupSignature_ x = Prelude.map pi13 <$> lookupSignature x
    where
	pi13 (x,y,z) = (x,z)

getParentSigs :: Name -> TC (Set.Set Name)
getParentSigs x =
    do	mcs <- Map.lookup x <$> getSignatures
	case mcs of
	    Just cs -> return $ sigParents cs
	    Nothing -> return $ Set.empty

isSubsignature :: Name -> Name -> TC Bool
isSubsignature x y =
    do	ps <- getParentSigs x
	return $ Set.member y ps

addSignature :: Name -> [Name] -> [(Name,FieldSource,DefValue)] -> TC ()
addSignature x ps cs =
    do	pps <- Set.unions <$> mapM getParentSigs ps
	let ps' = Set.union (Set.fromList ps) pps
	    sig = SigInfo cs ps'
	modify $ \s -> s { tcSignatures = insert x sig (tcSignatures s) }

delSignature :: Name -> TC ()
delSignature x =
    modify $ \s -> s { tcSignatures = delete x (tcSignatures s) }

setSignatures :: Signatures -> TC ()
setSignatures ds =
    modify $ \s -> s { tcSignatures = ds }

