-- Galois connections
-- Author :: Andreas Abel
-- Created:: Feb 15, 2005

-- # import Property

-- backward compatibility
Eq : (A:Set) -> A -> A -> Prop
Eq A a b = a == b

-- Relations

Rel : (X : Set) -> Type
Rel X = X -> X -> Prop

inv : (X : Set) -> Rel X -> Rel X
inv X R x y = R y x

-- Partial orders 

po : (X : Set) -> (R : Rel X) -> Prop
po X R =
  ( (x:X) -> (y:X) -> (R x y /\ R y x) <=> Eq X x y ) /\
  ( (x,y,z:X) -> R x y ==> R y z ==> R x z )

-- isotonicity

iso : (X : Set) -> (R : Rel X) -> (X -> X) -> Prop
iso X R f = (x,x':X) -> R x x' ==> R (f x) (f x')

-- inflationary (extensive) function

infl : (X : Set) -> (R : Rel X) -> (X -> X) -> Prop
infl X R f = (x:X) -> R x (f x)

-- deflationary (contractive) function

defl : (X : Set) -> (R : Rel X) -> (X -> X) -> Prop
defl X R f = infl X (inv X R) f

-- idempotent functions

idempot : (X : Set) -> (X -> X) -> Prop
idempot X f = (x:X) -> Eq X (f x) (f (f x))

-- closure operator

closOp : (X : Set) -> Rel X -> (X -> X) -> Prop
closOp X R f = iso X R f /\ infl X R f /\ idempot X f

-- Galois connection

galois : (X:Set) -> Rel X -> (Y:Set) -> Rel Y -> 
            (X -> Y) -> (Y -> X) -> Prop
galois X (<=) Y (=<) f g = (x:X) -> (y:Y) -> f x =< y <=> x <= g y

-- <= is a partial order on X

X    : Set
(<=) : Rel X
poX  : po X (<=)

-- =< is a partial order on Y

Y    : Set
(=<) : Rel Y
poY  : po Y (=<)

-- (f, g) is a Galois connection between X and Y

f : X -> Y
g : Y -> X
galFG : galois X (<=) Y (=<) f g

-- Theorem:: g o f is a closure operator

thm : closOp X (<=) (\ x -> g (f x))
thm = fol-plugin (poX,poY,galFG)

-- Impressive! Proved by Gandalf in no time!

