
-- # import nat
-- # import bool
-- # import id
-- # import prelude
-- # import eqchain

infixr 10 ::, ++

data Vec (A:Set) : Nat -> Set where
  nil	       : Vec A zero
  (::) (n:Nat) : A -> Vec A n -> Vec A (suc n)

-- head and tail

head (A:Set)(n:Nat) : Vec A (suc n) -> A
head (x :: xs) = x

tail (A:Set)(n:Nat) : Vec A (suc n) -> Vec A n
tail (x :: xs) = xs

-- append

(++) (A : Set)(n, m : Nat) :
     Vec A n -> Vec A m -> Vec A (n + m)
[]	  ++ ys	= ys
(x :: xs) ++ ys	= x :: (xs ++ ys)

-- split

mapFst (A, B, C : Set) : (f : A -> B) -> A × C -> B × C
mapFst f (pair x y) = pair (f x) y

split (A : Set) :
  (n, m : Nat) -> Vec A (n + m) ->
  Vec A n × Vec A m
split (suc n) m (x :: xs) = mapFst ((::) x) (split n m xs)
split 0	m xs		  = pair nil xs

-- split/append

splitAppend (A : Set)(n, m : Nat) :
	    (xs : Vec A n) -> (ys : Vec A m) ->
	    split n m (xs ++ ys) == pair xs ys
splitAppend []	      ys = refl
splitAppend (x :: xs) ys = rewrite splitAppend xs ys in refl

-- reverse

coerce (A : Set)(n, m : Nat) : n == m -> Vec A n -> Vec A m
coerce refl xs = xs

rev (A : Set)(n, m : Nat) : Vec A n -> Vec A m -> Vec A (n + m)
rev []	    ys	= ys
rev (x::xs) ys	= rev xs (x::ys)

reverse (A : Set)(n : Nat) : Vec A n -> Vec A n
reverse xs = rev xs []

-- vec

vec (A : Set)(n : Nat) : A -> Vec A n
vec {_} {0} x	  = []
vec {_} {suc n} x = x :: vec x

-- vapp
infixl 20 @@

(@@) (A, B : Set)(n : Nat) : Vec (A -> B) n -> Vec A n -> Vec B n
[]	@@ []	    = []
(f::fs) @@ (x::xs)  = f x :: fs @@ xs

-- map

map (A, B : Set)(n : Nat) : (A -> B) -> Vec A n -> Vec B n
map f xs  = vec f @@ xs

-- zip

zipWith (A, B, C : Set)(n : Nat) : (A -> B -> C) -> Vec A n -> Vec B n -> Vec C n
zipWith f xs ys = vec f @@ xs @@ ys

zip (A, B : Set)(n : Nat) : Vec A n -> Vec B n -> Vec (A × B) n
zip xs ys = zipWith pair xs ys

-- foldr

foldr (A, B : Set)(n : Nat) : (A -> B -> B) -> B -> Vec A n -> B
foldr f e []	  = e
foldr f e (x::xs) = f x (foldr f e xs)

-- sum

sum (n : Nat) : Vec Nat n -> Nat
sum = foldr (+) 0

-- count

boolValue : Bool -> Nat
boolValue true	= 1
boolValue false = 0

count (A : Set)(n : Nat) : (A -> Bool) -> Vec A n -> Nat
count p xs = sum $ map (boolValue @ p) xs

countLess (A : Set) :
	  (p : A -> Bool) -> (n : Nat) -> (xs : Vec A n) ->
	  count p xs <= n
countLess p 0	    []	    = le0
countLess p (suc n) (x::xs) = countL (p x) refl
  where
    countL : (b:Bool) -> p x == b -> count p (x::xs) <= suc n
    countL true  eq = rewrite eq in leS   (countLess p n xs)
    countL false eq = rewrite eq in leSuc (countLess p n xs)

{-
  if' (\b -> boolValue b + count p xs <= suc n)
      (p x)
      (leS (countLess p n xs))
      (leSuc (countLess p n xs))
-}

-- filter

mapmap (A,B,C:Set)(n : Nat) :
    (f : B -> C) -> (g : A -> B) -> (xs : Vec A n) ->
    map f (map g xs) == map (f @ g) xs
mapmap f g []	    = refl
mapmap f g (x::xs)  = cong (\zs -> f (g x) :: zs) (mapmap f g xs)

mapzip (A,B,C,D:Set)(n : Nat) :
      (f : C -> D) -> (g : A -> B -> C) -> (xs : Vec A n) -> (ys : Vec B n) ->
      map f (zipWith g xs ys) == zipWith (\x y -> f (g x y)) xs ys
mapzip f g [] []	    = refl
mapzip f g (x::xs) (y::ys)  = cong (\zs -> f (g x y) :: zs) (mapzip f g xs ys)

zipconst (A,B:Set)(n:Nat) :
	 (xs : Vec A n) -> (ys : Vec B n) ->
	 zipWith const xs ys == xs
zipconst [] []		  = refl
zipconst (x::xs) (y::ys)  = cong ((::) x) (zipconst xs ys)

filter' (A:Set)(n:Nat) : (xs : Vec (Bool × A) n) -> Vec A (count fst xs)
filter' []		     = []
filter' (pair true  x :: xs) = x :: filter' xs
filter' (pair false x :: xs) = filter' xs

filter2 (A : Set)(n : Nat) : (p : A -> Bool) -> (xs : Vec A n) -> Vec A (count p xs)
filter2 p xs =
  coerce (cong sum
	 $ eqChain (map (boolValue @ fst) (zip (map p xs) xs))

		=== map (boolValue @ fst) (zipWith (pair @ p) xs xs)

		      `since` cong (\zs -> map (boolValue @ fst) (zs @@ xs))
				   (mapmap pair p xs)

		=== zipWith (\x y -> boolValue (fst (pair (p x) y))) xs xs

		      `since` mapzip (boolValue @ fst) (pair @ p) xs xs

		=== map (boolValue @ p) (zipWith const xs xs)

		      `since` sym (mapzip (boolValue @ p) const xs xs)

		=== map (boolValue @ p) xs

		      `since` cong ((@@) (vec $ boolValue @ p)) (zipconst _ _)
	  )
	  $ filter' (zip (map p xs) xs)

filter (A : Set)(n : Nat) : (p : A -> Bool) -> (xs : Vec A n) -> Vec A (count p xs)
filter p [] = []
filter p (x::xs) with p x =>
       p (x::xs) true	= x :: filter p xs
       p (x::xs) false	= filter p xs

-- filter p (x::xs) with analyseBool (p x) =>
--        p (x::xs) (itsTrue  eq)	= rewrite eq in x :: filter p xs
--        p (x::xs) (itsFalse eq)	= rewrite eq in	     filter p xs

{-
-- testing
sequence : (n : Nat) -> Nat -> Vec Nat n
sequence 0	 m  = []
sequence (suc n) m  = m :: sequence n (suc m)

oneTo : (n : Nat) -> Vec Nat n
oneTo n = sequence n 1

test : (n : Nat) -> Vec Nat n × Vec Nat n
test n = split n n (xs ++ xs)
  where
    xs = oneTo n
-}
